# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.shape_calculator import _calculate_linear_regressor_output_shapes


def voting_regressor_shape_calculator(operator):
    return _calculate_linear_regressor_output_shapes(
        operator, enable_type_checking=False)


register_shape_calculator(
    'SklearnVotingRegressor', voting_regressor_shape_calculator)



# SPDX-License-Identifier: Apache-2.0


import numpy as np

from ..proto import onnx_proto
from ..common._apply_operation import (
    apply_cast, apply_concat, apply_reshape)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import guess_proto_type, Int64TensorType
from .._supported_operators import sklearn_operator_name_map


def _fetch_scores(scope, container, model, inputs, raw_scores=False,
                  is_regressor=False):
    op_type = sklearn_operator_name_map[type(model)]
    this_operator = scope.declare_local_operator(op_type, model)
    if container.has_options(model, 'raw_scores'):
        container.add_options(id(model), {'raw_scores': raw_scores})
    this_operator.inputs.append(inputs)
    if is_regressor:
        output_proba = scope.declare_local_variable(
            'variable', inputs.type.__class__())
        this_operator.outputs.append(output_proba)
    else:
        label_name = scope.declare_local_variable(
            'label', Int64TensorType())
        this_operator.outputs.append(label_name)
        output_proba = scope.declare_local_variable(
            'probability_tensor', inputs.type.__class__())
        this_operator.outputs.append(output_proba)

    proto_type = guess_proto_type(inputs.type)
    new_name = scope.get_unique_variable_name(
        output_proba.full_name + '_castio')
    apply_cast(scope, output_proba.full_name, new_name,
               container, to=proto_type)
    return new_name


def _add_passthrough_connection(operator, predictions):
    if operator.raw_operator.passthrough:
        predictions.append(operator.inputs[0].onnx_name)


def _transform_regressor(scope, operator, container, model):
    merged_prob_tensor = scope.get_unique_variable_name(
        'merged_probability_tensor')

    predictions = [
        _fetch_scores(
            scope, container, est, operator.inputs[0], is_regressor=True)
        for est in model.estimators_
    ]

    _add_passthrough_connection(operator, predictions)

    apply_concat(
        scope, predictions, merged_prob_tensor, container, axis=1)
    return merged_prob_tensor


def _transform(scope, operator, container, model):
    merged_prob_tensor = scope.get_unique_variable_name(
        'merged_probability_tensor')

    predictions = [
        _fetch_scores(scope, container, est, operator.inputs[0],
                      raw_scores=meth == 'decision_function')
        for est, meth in zip(model.estimators_, model.stack_method_)
        if est != 'drop'
    ]

    op = operator.raw_operator
    select_lact_column = (len(op.classes_) == 2 and all(
        op.stack_method_[est_idx] == 'predict_proba'
        for est_idx in range(0, len(op.estimators_))))
    if select_lact_column:
        column_index_name = scope.get_unique_variable_name('column_index')
        container.add_initializer(column_index_name,
                                  onnx_proto.TensorProto.INT64, [], [1])
        new_predictions = []
        for ipred, pred in enumerate(predictions):
            prob1 = scope.get_unique_variable_name('stack_prob%d' % ipred)
            container.add_node(
                'ArrayFeatureExtractor',
                [pred, column_index_name], prob1,
                name=scope.get_unique_operator_name('ArrayFeatureExtractor'),
                op_domain='ai.onnx.ml')
            new_predictions.append(prob1)
        predictions = new_predictions

    _add_passthrough_connection(operator, predictions)

    apply_concat(
        scope, predictions, merged_prob_tensor, container, axis=1)
    return merged_prob_tensor


def convert_sklearn_stacking_classifier(scope: Scope, operator: Operator,
                                        container: ModelComponentContainer):
    """
    Converter for StackingClassifier. It invokes converters for each
    estimator, concatenating their results before calling converter
    for the final estimator on the concatenated score.
    """
    stacking_op = operator.raw_operator
    classes = stacking_op.classes_
    options = container.get_options(stacking_op, dict(raw_scores=False))
    use_raw_scores = options['raw_scores']
    class_type = onnx_proto.TensorProto.STRING
    if (np.issubdtype(stacking_op.classes_.dtype, np.floating) or
            stacking_op.classes_.dtype == np.bool_):
        class_type = onnx_proto.TensorProto.INT32
        classes = classes.astype(np.int32)
    elif np.issubdtype(stacking_op.classes_.dtype, np.signedinteger):
        class_type = onnx_proto.TensorProto.INT32
    else:
        classes = np.array([s.encode('utf-8') for s in classes])

    classes_name = scope.get_unique_variable_name('classes')
    argmax_output_name = scope.get_unique_variable_name('argmax_output')
    reshaped_result_name = scope.get_unique_variable_name('reshaped_result')
    array_feature_extractor_result_name = scope.get_unique_variable_name(
        'array_feature_extractor_result')

    container.add_initializer(classes_name, class_type, classes.shape, classes)

    merged_proba_tensor = _transform(
        scope, operator, container, stacking_op)
    merge_proba = scope.declare_local_variable(
        'merged_stacked_proba', operator.inputs[0].type.__class__())
    container.add_node(
        'Identity', [merged_proba_tensor], [merge_proba.onnx_name])
    prob = _fetch_scores(
        scope, container, stacking_op.final_estimator_, merge_proba,
        raw_scores=use_raw_scores)
    container.add_node('Identity', prob, operator.outputs[1].onnx_name,
                       name=scope.get_unique_operator_name('OpProb'))
    container.add_node('ArgMax', prob,
                       argmax_output_name,
                       name=scope.get_unique_operator_name('ArgMax'), axis=1)
    container.add_node(
        'ArrayFeatureExtractor', [classes_name, argmax_output_name],
        array_feature_extractor_result_name, op_domain='ai.onnx.ml',
        name=scope.get_unique_operator_name('ArrayFeatureExtractor'))

    if class_type == onnx_proto.TensorProto.INT32:
        apply_reshape(scope, array_feature_extractor_result_name,
                      reshaped_result_name, container,
                      desired_shape=(-1,))
        apply_cast(scope, reshaped_result_name, operator.outputs[0].full_name,
                   container, to=onnx_proto.TensorProto.INT64)
    else:
        apply_reshape(scope, array_feature_extractor_result_name,
                      operator.outputs[0].full_name, container,
                      desired_shape=(-1,))


def convert_sklearn_stacking_regressor(scope: Scope, operator: Operator,
                                       container: ModelComponentContainer):
    """
    Converter for StackingRegressor. It invokes converters for each
    estimator, concatenating their results before calling converter
    for the final estimator on the concatenated score.
    """
    stacking_op = operator.raw_operator

    merged_proba_tensor = _transform_regressor(
        scope, operator, container, stacking_op)
    merge_proba = scope.declare_local_variable(
        'merged_stacked_proba', operator.inputs[0].type.__class__())
    container.add_node(
        'Identity', [merged_proba_tensor], [merge_proba.onnx_name])
    prob = _fetch_scores(
        scope, container, stacking_op.final_estimator_, merge_proba,
        is_regressor=True)
    container.add_node('Identity', prob, operator.outputs[0].full_name,
                       name=scope.get_unique_operator_name('Identity'))


register_converter('SklearnStackingClassifier',
                   convert_sklearn_stacking_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'nocl': [True, False],
                            'output_class_labels': [False, True],
                            'raw_scores': [True, False]})
register_converter('SklearnStackingRegressor',
                   convert_sklearn_stacking_regressor)



# SPDX-License-Identifier: Apache-2.0

from .onnx_operator import OnnxOperator


class OnnxCDist(OnnxOperator):
    """
    Defines a custom operator not defined by ONNX
    specifications but in onnxruntime.
    """

    since_version = 1
    expected_inputs = [('X', 'T'), ('Y', 'T')]
    expected_outputs = [('dist', 'T')]
    input_range = [2, 2]
    output_range = [1, 1]
    is_deprecated = False
    domain = 'com.microsoft'
    operator_name = 'CDist'
    past_version = {}

    def __init__(self, X, Y, metric='sqeuclidean', op_version=None,
                 **kwargs):
        """
        :param X: array or OnnxOperatorMixin
        :param Y: array or OnnxOperatorMixin
        :param metric: distance type
        :param dtype: *np.float32* or *np.float64*
        :param op_version: opset version
        :param kwargs: addition parameter
        """
        OnnxOperator.__init__(self, X, Y, metric=metric,
                              op_version=op_version, **kwargs)


class OnnxSolve(OnnxOperator):
    """
    Defines a custom operator not defined by ONNX
    specifications but in onnxruntime.
    """

    since_version = 1
    expected_inputs = [('A', 'T'), ('Y', 'T')]
    expected_outputs = [('X', 'T')]
    input_range = [2, 2]
    output_range = [1, 1]
    is_deprecated = False
    domain = 'com.microsoft'
    operator_name = 'Solve'
    past_version = {}

    def __init__(self, A, Y, lower=False, transposed=False,
                 op_version=None, **kwargs):
        """
        :param A: array or OnnxOperatorMixin
        :param Y: array or OnnxOperatorMixin
        :param lower: see :epkg:`solve`
        :param transposed: see :epkg:`solve`
        :param dtype: *np.float32* or *np.float64*
        :param op_version: opset version
        :param kwargs: additional parameters
        """
        OnnxOperator.__init__(self, A, Y,
                              lower=lower, transposed=transposed,
                              op_version=op_version, **kwargs)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from onnx import TensorProto
from onnx.helper import make_tensor
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_feature_hasher(scope: Scope, operator: Operator,
                                   container: ModelComponentContainer):
    X = operator.inputs[0]
    out = operator.outputs
    op = operator.raw_operator
    if op.input_type != "string":
        raise RuntimeError(
            f"The converter for FeatureHasher only supports "
            f"input_type='string' not {op.input_type!r}.")

    hashed_ = scope.get_unique_variable_name('hashed_')
    container.add_node('MurmurHash3', X.full_name, hashed_,
                       positive=0, seed=0, op_domain="com.microsoft",
                       op_version=1)
    hashed = scope.get_unique_variable_name('hashed')
    container.add_node('Cast', hashed_, hashed, to=TensorProto.INT64)

    if op.dtype in (np.float32, np.float64, np.int64):
        cst_neg = -1
    else:
        cst_neg = 4294967295

    infinite = scope.get_unique_variable_name('infinite')
    container.add_initializer(infinite, TensorProto.INT64, [1],
                              [-2147483648])

    infinite2 = scope.get_unique_variable_name('infinite2')
    container.add_initializer(infinite2, TensorProto.INT64, [1],
                              [cst_neg])

    infinite_n = scope.get_unique_variable_name('infinite_n')
    container.add_initializer(infinite_n, TensorProto.INT64, [1],
                              [2147483647 - (op.n_features - 1)])

    zero = scope.get_unique_variable_name('zero')
    container.add_initializer(zero, TensorProto.INT64, [1], [0])

    one = scope.get_unique_variable_name('one')
    container.add_initializer(one, TensorProto.INT64, [1], [1])

    mone = scope.get_unique_variable_name('mone')
    container.add_initializer(mone, TensorProto.INT64, [1], [-1])

    mtwo = scope.get_unique_variable_name('mtwo')
    container.add_initializer(mtwo, TensorProto.INT64, [1], [-2])

    nf = scope.get_unique_variable_name('nf')
    container.add_initializer(nf, TensorProto.INT64, [1], [op.n_features])

    new_shape = scope.get_unique_variable_name('new_shape')
    container.add_initializer(new_shape, TensorProto.INT64, [2], [-1, 1])
    new_shape2 = scope.get_unique_variable_name('new_shape2')
    container.add_initializer(new_shape2, TensorProto.INT64, [2], [1, -1])

    # values
    if op.alternate_sign:
        cmp = scope.get_unique_variable_name('cmp')
        container.add_node('GreaterOrEqual', [hashed, zero], cmp)
        values = scope.get_unique_variable_name('values')
        container.add_node('Where', [cmp, one, infinite2], values)
    else:
        mul = scope.get_unique_variable_name('mul')
        container.add_node('Mul', [hashed, zero], mul)
        values = scope.get_unique_variable_name('values')
        container.add_node('Add', [mul, one], values)

    values_reshaped = scope.get_unique_variable_name('values_reshaped')
    container.add_node('Reshape', [values, new_shape], values_reshaped)

    # indices
    cmp = scope.get_unique_variable_name('cmp_ind')
    container.add_node('Equal', [hashed, infinite], cmp)
    values_abs = scope.get_unique_variable_name('values_abs')
    container.add_node('Abs', hashed, values_abs)
    values_ind = scope.get_unique_variable_name('values_ind')
    container.add_node('Where', [cmp, infinite_n, values_abs], values_ind)
    indices = scope.get_unique_variable_name('indices')
    container.add_node('Mod', [values_ind, nf], indices)
    indices_reshaped = scope.get_unique_variable_name('indices_reshaped')
    container.add_node('Reshape', [indices, new_shape], indices_reshaped)

    # scatter
    zerot_ = scope.get_unique_variable_name('zerot_')
    container.add_node('ConstantOfShape', [nf], zerot_,
                       value=make_tensor("value",
                                         TensorProto.INT64, [1], [0]))
    zerot = scope.get_unique_variable_name('zerot')
    container.add_node('Mul', [indices_reshaped, zerot_], zerot)

    final = scope.get_unique_variable_name('final')
    container.add_node('ScatterElements',
                       [zerot, indices_reshaped, values_reshaped],
                       final, axis=1)

    # at this point, every string has been processed as if it was in
    # in a single columns.
    # in case there is more than one column, we need to reduce over
    # the last dimension
    input_shape = scope.get_unique_variable_name('input_shape')
    container.add_node('Shape', X.full_name, input_shape)
    shape_not_last = scope.get_unique_variable_name('shape_not_last')
    container.add_node('Slice', [input_shape, zero, mone], shape_not_last)
    final_shape = scope.get_unique_variable_name('final_last')
    container.add_node('Concat', [shape_not_last, mone, nf],
                       final_shape, axis=0)
    final_reshaped = scope.get_unique_variable_name('final_reshaped')
    container.add_node('Reshape', [final, final_shape], final_reshaped)
    container.add_node('ReduceSum', [final_reshaped, mtwo],
                       out[0].full_name, keepdims=0)


register_converter('SklearnFeatureHasher', convert_sklearn_feature_hasher)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import guess_proto_type


def convert_sklearn_replace_transformer(scope: Scope, operator: Operator,
                                        container: ModelComponentContainer):
    op = operator.raw_operator
    input_name = operator.inputs[0].full_name
    output_name = operator.outputs[0].full_name

    proto_dtype = guess_proto_type(operator.inputs[0].type)

    cst_nan_name = scope.get_unique_variable_name('nan_name')
    container.add_initializer(cst_nan_name, proto_dtype, [1], [op.to_value])
    cst_zero_name = scope.get_unique_variable_name('zero_name')
    container.add_initializer(
        cst_zero_name, proto_dtype, [1], [op.from_value])

    mask_name = scope.get_unique_variable_name('mask_name')
    container.add_node('Equal', [input_name, cst_zero_name],
                       mask_name,
                       name=scope.get_unique_operator_name('Equal'))

    container.add_node('Where', [mask_name, cst_nan_name, input_name],
                       output_name,
                       name=scope.get_unique_operator_name('Where'))


register_converter(
    'SklearnReplaceTransformer', convert_sklearn_replace_transformer)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ._apply_operation import apply_cast, apply_reshape
from ..proto import onnx_proto


def get_label_classes(scope, op, node_names=False):
    """
    Extracts the model classes,
    handles option ``nocl`` and ``zipmap=='columns'``
    """
    options = scope.get_options(op, dict(nocl=False))
    if options['nocl']:
        if len(op.classes_.shape) > 1 and op.classes_.shape[1] > 1:
            raise RuntimeError(
                "Options 'nocl=True' is not implemented for multi-label "
                "classification (class: {}).".format(op.__class__.__name__))
        classes = np.arange(0, len(op.classes_))
    elif node_names:
        try:
            options = scope.get_options(op, dict(zipmap=False))
            zipcol = options['zipmap'] == 'columns'
        except NameError:
            zipcol = False
        if zipcol:
            clnames = op.classes_.ravel()
            if (np.issubdtype(clnames.dtype, np.integer) or
                    clnames.dtype == np.bool_):
                classes = np.array(['i%d' % c for c in clnames])
            else:
                classes = np.array(['s%s' % c for c in clnames])
        else:
            classes = op.classes_
    elif hasattr(op, 'classes_'):
        classes = op.classes_
    elif hasattr(op, 'intercept_'):
        classes = len(op.intercept_)
    else:
        raise RuntimeError(
            "No known ways to retrieve the number of classes for class %r."
            "" % type(op))
    return classes


def _finalize_converter_classes(scope, argmax_output_name, output_full_name,
                                container, classes, proto_dtype):
    """
    See :func:`convert_voting_classifier`.
    """
    if (np.issubdtype(classes.dtype, np.floating) or
            classes.dtype == np.bool_):
        class_type = onnx_proto.TensorProto.INT32
        classes = np.array(list(map(lambda x: int(x), classes)))
    elif np.issubdtype(classes.dtype, np.signedinteger):
        class_type = onnx_proto.TensorProto.INT32
    else:
        classes = np.array([s.encode('utf-8') for s in classes])
        class_type = onnx_proto.TensorProto.STRING

    classes_name = scope.get_unique_variable_name('classes')
    container.add_initializer(classes_name, class_type, classes.shape, classes)

    array_feature_extractor_result_name = scope.get_unique_variable_name(
        'array_feature_extractor_result')
    container.add_node(
        'ArrayFeatureExtractor', [classes_name, argmax_output_name],
        array_feature_extractor_result_name, op_domain='ai.onnx.ml',
        name=scope.get_unique_operator_name('ArrayFeatureExtractor'))

    output_shape = (-1,)
    if class_type == onnx_proto.TensorProto.INT32:
        cast2_result_name = scope.get_unique_variable_name('cast2_result')
        reshaped_result_name = scope.get_unique_variable_name(
            'reshaped_result')
        apply_cast(scope, array_feature_extractor_result_name,
                   cast2_result_name, container,
                   to=proto_dtype)
        apply_reshape(scope, cast2_result_name, reshaped_result_name,
                      container, desired_shape=output_shape)
        apply_cast(scope, reshaped_result_name, output_full_name, container,
                   to=onnx_proto.TensorProto.INT64)
    else:  # string labels
        apply_reshape(scope, array_feature_extractor_result_name,
                      output_full_name, container, desired_shape=output_shape)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..algebra.onnx_ops import (
    OnnxConcat, OnnxReshapeApi13, OnnxIdentity)
try:
    from ..algebra.onnx_ops import OnnxSequenceConstruct
except ImportError:
    # Available since opset 11
    OnnxSequenceConstruct = None
from ..algebra.onnx_operator import OnnxSubEstimator


def convert_multi_output_regressor_converter(
        scope: Scope, operator: Operator, container: ModelComponentContainer):
    """
    Converts a *MultiOutputRegressor* into *ONNX* format.
    """
    op_version = container.target_opset
    op = operator.raw_operator
    inp = operator.inputs[0]
    y_list = [
        OnnxReshapeApi13(
            OnnxSubEstimator(sub, inp, op_version=op_version),
            np.array([-1, 1], dtype=np.int64),
            op_version=op_version)
        for sub in op.estimators_]

    output = OnnxConcat(*y_list, axis=1, op_version=op_version,
                        output_names=[operator.outputs[0]])
    output.add_to(scope=scope, container=container)


def convert_multi_output_classifier_converter(
        scope: Scope, operator: Operator, container: ModelComponentContainer):
    """
    Converts a *MultiOutputClassifier* into *ONNX* format.
    """
    if OnnxSequenceConstruct is None:
        raise RuntimeError(
            "This converter requires opset>=11.")
    op_version = container.target_opset
    op_version = container.target_opset
    op = operator.raw_operator
    inp = operator.inputs[0]
    options = scope.get_options(op)
    if options.get('nocl', True):
        options = options.copy()
    else:
        options = {}

    options.update({'zipmap': False})
    y_list = [OnnxSubEstimator(sub, inp, op_version=op_version,
                               options=options)
              for sub in op.estimators_]

    # labels
    label_list = [OnnxReshapeApi13(y[0], np.array([-1, 1], dtype=np.int64),
                                   op_version=op_version)
                  for y in y_list]

    # probabilities
    proba_list = [OnnxIdentity(y[1], op_version=op_version)
                  for y in y_list]
    label = OnnxConcat(*label_list, axis=1, op_version=op_version,
                       output_names=[operator.outputs[0]])
    label.add_to(scope=scope, container=container)

    proba = OnnxSequenceConstruct(
        *proba_list, op_version=op_version,
        output_names=[operator.outputs[1]])
    proba.add_to(scope=scope, container=container)


register_converter('SklearnMultiOutputRegressor',
                   convert_multi_output_regressor_converter)
register_converter('SklearnMultiOutputClassifier',
                   convert_multi_output_classifier_converter,
                   options={'nocl': [False, True],
                            'output_class_labels': [False, True],
                            'zipmap': [False, True]})



# SPDX-License-Identifier: Apache-2.0


import warnings
import logging

# Calibrated classifier CV
from sklearn.calibration import CalibratedClassifierCV

# Linear classifiers
from sklearn.linear_model import (
    LogisticRegression, LogisticRegressionCV,
    PassiveAggressiveClassifier,
    Perceptron,
    RidgeClassifier, RidgeClassifierCV,
    SGDClassifier
)
from sklearn.svm import LinearSVC, OneClassSVM

# Linear regressors
from sklearn.linear_model import (
    ARDRegression,
    BayesianRidge,
    ElasticNet, ElasticNetCV,
    HuberRegressor,
    Lars, LarsCV,
    Lasso, LassoCV,
    LassoLars, LassoLarsCV,
    LassoLarsIC,
    LinearRegression,
    MultiTaskElasticNet, MultiTaskElasticNetCV,
    MultiTaskLasso, MultiTaskLassoCV,
    OrthogonalMatchingPursuit, OrthogonalMatchingPursuitCV,
    PassiveAggressiveRegressor,
    RANSACRegressor,
    Ridge, RidgeCV,
    SGDRegressor,
    TheilSenRegressor
)
try:
    from sklearn.linear_model import GammaRegressor
except ImportError:
    # available since sklearn>=1.1
    GammaRegressor = None
try:
    from sklearn.linear_model import QuantileRegressor
except ImportError:
    # available since sklearn>=1.0
    QuantileRegressor = None
try:
    from sklearn.linear_model import PoissonRegressor
except ImportError:
    # available since sklearn>=0.23
    PoissonRegressor = None
try:
    from sklearn.linear_model import TweedieRegressor
except ImportError:
    # available since sklearn>=0.23
    TweedieRegressor = None
try:
    from sklearn.linear_model import SGDOneClassSVM
except ImportError:
    # available since sklearn>=1.0
    SGDOneClassSVM = None

from sklearn.svm import LinearSVR
from sklearn.discriminant_analysis import (
    LinearDiscriminantAnalysis,
    QuadraticDiscriminantAnalysis
)

# Mixture
from sklearn.mixture import (
    GaussianMixture, BayesianGaussianMixture
)

# Multi-class
from sklearn.multiclass import OneVsRestClassifier, OneVsOneClassifier

# Tree-based models
from sklearn.ensemble import (
    AdaBoostClassifier, AdaBoostRegressor,
    BaggingClassifier, BaggingRegressor,
    ExtraTreesClassifier, ExtraTreesRegressor,
    GradientBoostingClassifier, GradientBoostingRegressor,
    IsolationForest,
    RandomForestClassifier, RandomForestRegressor, RandomTreesEmbedding,
    VotingClassifier
)
try:
    from sklearn.ensemble import VotingRegressor
except ImportError:
    # New in 0.21
    VotingRegressor = None
try:
    from sklearn.ensemble import StackingClassifier, StackingRegressor
except ImportError:
    # New in 0.22
    StackingClassifier = None
    StackingRegressor = None
from sklearn.tree import (
    DecisionTreeClassifier, DecisionTreeRegressor,
    ExtraTreeClassifier, ExtraTreeRegressor
)

# Gaussian processes
from sklearn.gaussian_process import (
    GaussianProcessClassifier, GaussianProcessRegressor
)

# GridSearchCV
from sklearn.model_selection import GridSearchCV

# MultiOutput
from sklearn.multioutput import MultiOutputClassifier, MultiOutputRegressor

# Support vector machines
from sklearn.svm import NuSVC, NuSVR, SVC, SVR

# K-nearest neighbors
from sklearn.neighbors import (
    KNeighborsClassifier,
    KNeighborsRegressor,
    LocalOutlierFactor,
    NearestNeighbors,
    RadiusNeighborsClassifier,
    RadiusNeighborsRegressor,
)
try:
    from sklearn.neighbors import (
        KNeighborsTransformer,
        NeighborhoodComponentsAnalysis,
    )
except ImportError:
    # New in 0.22
    KNeighborsTransformer = None
    NeighborhoodComponentsAnalysis = None

# Naive Bayes
from sklearn.naive_bayes import (
    BernoulliNB,
    GaussianNB,
    MultinomialNB,
)
try:
    from sklearn.naive_bayes import CategoricalNB
except ImportError:
    # scikit-learn versions <= 0.21
    CategoricalNB = None
try:
    from sklearn.naive_bayes import ComplementNB
except ImportError:
    # scikit-learn versions <= 0.19
    ComplementNB = None

# Neural Networks
from sklearn.neural_network import MLPClassifier, MLPRegressor

# Clustering
from sklearn.cluster import KMeans, MiniBatchKMeans

# Operators for preprocessing and feature engineering
from sklearn.cross_decomposition import PLSRegression
from sklearn.decomposition import (
    KernelPCA,
    IncrementalPCA,
    PCA,
    TruncatedSVD,
)
from sklearn.feature_extraction import (
    DictVectorizer,
    FeatureHasher,
)
from sklearn.feature_extraction.text import (
    CountVectorizer, TfidfTransformer, TfidfVectorizer
)
from sklearn.feature_selection import (
    GenericUnivariateSelect, RFE, RFECV,
    SelectFdr, SelectFpr, SelectFromModel,
    SelectFwe, SelectKBest, SelectPercentile,
    VarianceThreshold
)
try:
    # 0.20
    from sklearn.impute import SimpleImputer
except ImportError:
    # 0.19
    from sklearn.preprocessing import Imputer as SimpleImputer
from sklearn.preprocessing import Binarizer
try:
    from sklearn.preprocessing import Imputer
except ImportError:
    # removed in 0.21
    Imputer = None
try:
    from sklearn.impute import KNNImputer
except ImportError:
    # New in 0.22
    KNNImputer = None
try:
    from sklearn.preprocessing import KBinsDiscretizer
except ImportError:
    # not available in 0.19
    KBinsDiscretizer = None
from sklearn.preprocessing import (
    LabelBinarizer, LabelEncoder,
    Normalizer, OneHotEncoder
)
try:
    from sklearn.preprocessing import OrdinalEncoder
except ImportError:
    # Not available in scikit-learn < 0.20.0
    OrdinalEncoder = None
from sklearn.preprocessing import (
    FunctionTransformer,
    KernelCenterer,
    MaxAbsScaler,
    MinMaxScaler,
    PolynomialFeatures,
    RobustScaler,
    StandardScaler
)

try:
    from sklearn.preprocessing import PowerTransformer
except ImportError:
    # Not available in scikit-learn < 0.20.0
    PowerTransformer = None

try:
    from sklearn.ensemble import (
        HistGradientBoostingClassifier,
        HistGradientBoostingRegressor
    )
except ImportError:
    # Second verification as these models still require
    # manual activation.
    try:
        from sklearn.ensemble._hist_gradient_boosting.gradient_boosting import (  # noqa
            HistGradientBoostingClassifier,
            HistGradientBoostingRegressor
        )
    except ImportError:
        HistGradientBoostingRegressor = None
        HistGradientBoostingClassifier = None

from sklearn.random_projection import GaussianRandomProjection

try:
    from sklearn.compose import ColumnTransformer
except ImportError:
    # ColumnTransformer was introduced in 0.20.
    ColumnTransformer = None

from sklearn.pipeline import Pipeline, FeatureUnion

# Custom extension
from .sklapi import CastRegressor, CastTransformer, ReplaceTransformer

from .common._registration import register_converter, register_shape_calculator

logger = logging.getLogger('skl2onnx')

# In most cases, scikit-learn operator produces only one output.
# However, each classifier has basically two outputs; one is the
# predicted label and the other one is the probabilities of all
# possible labels. Here is a list of supported scikit-learn
# classifiers. In the parsing stage, we produce two outputs for objects
# included in the following list and one output for everything not in
# the list.
sklearn_classifier_list = list(filter(lambda m: m is not None, [
    AdaBoostClassifier,
    BaggingClassifier,
    BernoulliNB,
    CategoricalNB,
    CalibratedClassifierCV,
    ComplementNB,
    DecisionTreeClassifier,
    ExtraTreeClassifier,
    ExtraTreesClassifier,
    GaussianNB,
    GaussianProcessClassifier,
    GradientBoostingClassifier,
    HistGradientBoostingClassifier,
    KNeighborsClassifier,
    LinearDiscriminantAnalysis,
    LinearSVC,
    LogisticRegression,
    LogisticRegressionCV,
    MLPClassifier,
    MultinomialNB,
    NuSVC,
    OneVsOneClassifier,
    OneVsRestClassifier,
    PassiveAggressiveClassifier,
    Perceptron,
    QuadraticDiscriminantAnalysis,
    RandomForestClassifier,
    SGDClassifier,
    StackingClassifier,
    SVC,
    VotingClassifier,
]))

# Clustering algorithms: produces two outputs, label and score for
# each cluster in most cases.
cluster_list = [KMeans, MiniBatchKMeans]

# Outlier detection algorithms:
# produces two outputs, label and scores
outlier_list = [IsolationForest, LocalOutlierFactor, OneClassSVM]


# Associate scikit-learn types with our operator names. If two
# scikit-learn models share a single name, it means their are
# equivalent in terms of conversion.
def build_sklearn_operator_name_map():
    res = {k: "Sklearn" + k.__name__ for k in [
        AdaBoostClassifier,
        AdaBoostRegressor,
        BaggingClassifier,
        BaggingRegressor,
        BayesianGaussianMixture,
        BayesianRidge,
        BernoulliNB,
        Binarizer,
        CalibratedClassifierCV,
        CategoricalNB,
        CastRegressor,
        CastTransformer,
        ColumnTransformer,
        ComplementNB,
        CountVectorizer,
        DictVectorizer,
        DecisionTreeClassifier,
        DecisionTreeRegressor,
        ExtraTreeClassifier,
        ExtraTreeRegressor,
        ExtraTreesClassifier,
        ExtraTreesRegressor,
        FeatureHasher,
        FeatureUnion,
        FunctionTransformer,
        GammaRegressor,
        GaussianNB,
        GaussianMixture,
        GaussianProcessClassifier,
        GaussianProcessRegressor,
        GaussianRandomProjection,
        GenericUnivariateSelect,
        GradientBoostingClassifier,
        GradientBoostingRegressor,
        HistGradientBoostingClassifier,
        HistGradientBoostingRegressor,
        Imputer,
        IncrementalPCA,
        IsolationForest,
        KMeans,
        LabelBinarizer,
        LabelEncoder,
        LinearRegression,
        LinearSVC,
        LinearSVR,
        LocalOutlierFactor,
        MaxAbsScaler,
        MiniBatchKMeans,
        MinMaxScaler,
        MLPClassifier,
        MLPRegressor,
        MultinomialNB,
        MultiOutputClassifier,
        MultiOutputRegressor,
        KBinsDiscretizer,
        KernelCenterer,
        KernelPCA,
        KNeighborsClassifier,
        KNeighborsRegressor,
        KNeighborsTransformer,
        KNNImputer,
        NearestNeighbors,
        NeighborhoodComponentsAnalysis,
        Normalizer,
        OneClassSVM,
        OneHotEncoder,
        OneVsOneClassifier,
        OneVsRestClassifier,
        OrdinalEncoder,
        PCA,
        PLSRegression,
        Pipeline,
        PoissonRegressor,
        PolynomialFeatures,
        PowerTransformer,
        QuadraticDiscriminantAnalysis,
        RadiusNeighborsClassifier,
        RadiusNeighborsRegressor,
        RandomForestClassifier,
        RandomForestRegressor,
        RandomTreesEmbedding,
        RANSACRegressor,
        ReplaceTransformer,
        RFE,
        RFECV,
        RobustScaler,
        SelectFdr,
        SelectFpr,
        SelectFromModel,
        SelectFwe,
        SelectKBest,
        SelectPercentile,
        SGDClassifier,
        SGDOneClassSVM,
        SimpleImputer,
        StackingClassifier,
        StackingRegressor,
        SVC,
        SVR,
        TfidfVectorizer,
        TfidfTransformer,
        TruncatedSVD,
        TweedieRegressor,
        VarianceThreshold,
        VotingClassifier,
        VotingRegressor,
    ] if k is not None}
    res.update({
        ARDRegression: 'SklearnLinearRegressor',
        ElasticNet: 'SklearnLinearRegressor',
        ElasticNetCV: 'SklearnLinearRegressor',
        GridSearchCV: 'SklearnGridSearchCV',
        HuberRegressor: 'SklearnLinearRegressor',
        LinearRegression: 'SklearnLinearRegressor',
        Lars: 'SklearnLinearRegressor',
        LarsCV: 'SklearnLinearRegressor',
        Lasso: 'SklearnLinearRegressor',
        LassoCV: 'SklearnLinearRegressor',
        LassoLars: 'SklearnLinearRegressor',
        LassoLarsCV: 'SklearnLinearRegressor',
        LassoLarsIC: 'SklearnLinearRegressor',
        LinearDiscriminantAnalysis: 'SklearnLinearClassifier',
        LogisticRegression: 'SklearnLinearClassifier',
        LogisticRegressionCV: 'SklearnLinearClassifier',
        MultiTaskElasticNet: 'SklearnLinearRegressor',
        MultiTaskElasticNetCV: 'SklearnLinearRegressor',
        MultiTaskLasso: 'SklearnLinearRegressor',
        MultiTaskLassoCV: 'SklearnLinearRegressor',
        NuSVC: 'SklearnSVC',
        NuSVR: 'SklearnSVR',
        OrthogonalMatchingPursuit: 'SklearnLinearRegressor',
        OrthogonalMatchingPursuitCV: 'SklearnLinearRegressor',
        PassiveAggressiveClassifier: 'SklearnSGDClassifier',
        PassiveAggressiveRegressor: 'SklearnLinearRegressor',
        Perceptron: 'SklearnSGDClassifier',
        QuantileRegressor: 'SklearnLinearRegressor',
        Ridge: 'SklearnLinearRegressor',
        RidgeCV: 'SklearnLinearRegressor',
        RidgeClassifier: 'SklearnLinearClassifier',
        RidgeClassifierCV: 'SklearnLinearClassifier',
        SGDRegressor: 'SklearnLinearRegressor',
        StandardScaler: 'SklearnScaler',
        TheilSenRegressor: 'SklearnLinearRegressor',
    })
    if None in res:
        del res[None]
    return res


def update_registered_converter(model, alias, shape_fct, convert_fct,
                                overwrite=True, parser=None, options=None):
    """
    Registers or updates a converter for a new model so that
    it can be converted when inserted in a *scikit-learn* pipeline.

    :param model: model class
    :param alias: alias used to register the model
    :param shape_fct: function which checks or modifies the expected
        outputs, this function should be fast so that the whole graph
        can be computed followed by the conversion of each model,
        parallelized or not
    :param convert_fct: function which converts a model
    :param overwrite: False to raise exception if a converter
        already exists
    :param parser: overwrites the parser as well if not empty
    :param options: registered options for this converter

    The alias is usually the library name followed by the model name.
    Example:

    ::

        from skl2onnx.common.shape_calculator import calculate_linear_classifier_output_shapes
        from skl2onnx.operator_converters.RandomForest import convert_sklearn_random_forest_classifier
        from skl2onnx import update_registered_converter
        update_registered_converter(
                SGDClassifier, 'SklearnLinearClassifier',
                calculate_linear_classifier_output_shapes,
                convert_sklearn_random_forest_classifier,
                options={'zipmap': [True, False, 'columns'],
                         'output_class_labels': [False, True],
                         'raw_scores': [True, False]})

    The function does not update the parser if not specified except if
    option `'zipmap'` is added to the list. Every classifier
    must declare this option to let the default parser
    automatically handle that option.
    """  # noqa
    if (not overwrite and model in sklearn_operator_name_map
            and alias != sklearn_operator_name_map[model]):
        warnings.warn("Model '{0}' was already registered under alias "
                      "'{1}'.".format(model, sklearn_operator_name_map[model]))
    sklearn_operator_name_map[model] = alias
    register_converter(alias, convert_fct, overwrite=overwrite,
                       options=options)
    register_shape_calculator(alias, shape_fct, overwrite=overwrite)
    if parser is not None:
        from ._parse import update_registered_parser
        update_registered_parser(model, parser)
    elif (options is not None and
            ('zipmap' in options or 'output_class_labels' in options)):
        from ._parse import (
            _parse_sklearn_classifier, update_registered_parser)
        update_registered_parser(model, _parse_sklearn_classifier)


def _get_sklearn_operator_name(model_type):
    """
    Get operator name of the input argument

    :param model_type:  A scikit-learn object (e.g., SGDClassifier
                        and Binarizer)
    :return: A string which stands for the type of the input model in
             our conversion framework
    """
    if model_type not in sklearn_operator_name_map:
        # No proper operator name found, it means a local operator.
        alias = None
    else:
        alias = sklearn_operator_name_map[model_type]
    logger.debug('[parsing] found alias=%r for type=%r.', alias, model_type)
    return alias


def get_model_alias(model_type):
    """
    Get alias model. Raise an exception if not found.

    :param model_type:  A scikit-learn object (e.g., SGDClassifier
                        and Binarizer)
    :return: A string which stands for the type of the input model in
             our conversion framework
    """
    res = _get_sklearn_operator_name(model_type)
    if res is None:
        raise RuntimeError("Unable to find alias for model '{}'. "
                           "The converter is likely missing."
                           "".format(model_type))
    return res


# registered converters
sklearn_operator_name_map = build_sklearn_operator_name_map()



# SPDX-License-Identifier: Apache-2.0


import warnings
from collections import OrderedDict, Counter
import numpy as np
from ..common._apply_operation import (
    apply_cast, apply_reshape, apply_identity)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import guess_proto_type, StringTensorType
from ..proto import onnx_proto
from ..algebra.onnx_ops import OnnxStringNormalizer


def _intelligent_split(text, op, tokenizer, existing):
    """
    Splits text into tokens. *scikit-learn*
    merges tokens with ``' '.join(tokens)``
    to name ngrams. ``'a  b'`` could be ``('a ', 'b')``
    or ``('a', ' b')``.
    See `ngram sequence
    <https://github.com/scikit-learn/scikit-learn/blob/master/
    sklearn/feature_extraction/text.py#L169>`_.
    """
    if op.analyzer == 'word':
        if op.ngram_range[0] == op.ngram_range[1] == 1:
            spl = [text]
        elif op.ngram_range[0] == 1 and len(text) >= 2:
            # Every element is in the vocabulary.
            # Naive method
            p1 = len(text) - len(text.lstrip())
            p2_ = len(text) - len(text.rstrip())
            if p2_ == 0:
                p2 = len(text)
            else:
                p2 = -p2_
            spl = text[p1:p2].split()
            if len(spl) <= 1:
                spl = [text]
            else:
                spl[0] = " " * p1 + spl[0]
                spl[-1] = spl[-1] + " " * p2_
            exc = None
            if len(spl) == 1:
                pass
            elif len(spl) == 2:
                if (spl[0] not in op.vocabulary_ or
                        spl[1] not in op.vocabulary_):
                    # This is neceassarily a single token.
                    spl = [text]
                elif spl[0] in op.vocabulary_ and spl[1] in op.vocabulary_:
                    # ambiguity
                    # w1, w2 can be either a 2-grams, either a token.
                    # Usually, ' ' is not part of any token.
                    pass
            elif len(spl) == 3:
                stok = (all([s in op.vocabulary_ for s in spl]), spl)
                spl12 = (spl[2] in op.vocabulary_ and
                         (spl[0] + ' ' + spl[1]) in op.vocabulary_,
                         [spl[0] + ' ' + spl[1], spl[2]])
                spl23 = (spl[0] in op.vocabulary_ and
                         (spl[1] + ' ' + spl[2]) in op.vocabulary_,
                         [spl[0], spl[1] + ' ' + spl[2]])
                c = Counter(map(lambda t: t[0], [stok, spl12, spl23]))
                if c.get(True, -1) == 0:
                    spl = [text]
                found = [el[1] for el in [stok, spl12, spl23] if el[0]]
                if len(found) == 1:
                    spl = found[0]
                elif len(found) == 0:
                    spl = [text]
                elif stok[0]:
                    # By default, we assume the token is just the sum of
                    # single words.
                    pass
                else:
                    exc = (
                        "More than one decomposition in tokens: [" +
                        ", ".join(map(lambda t: "-".join(t), found)) + "].")
            elif any(map(lambda g: g in op.vocabulary_, spl)):
                # TODO: handle this case with an algorithm
                # which is able to break a string into
                # known substrings.
                exc = "Unable to identify tokens in n-grams."
            if exc:
                raise RuntimeError(
                    "Unable to split n-grams '{}' into tokens. "
                    "{} This happens when a token contain "
                    "spaces. Token '{}' may be a token or a n-gram '{}'."
                    "".format(text, exc, text, spl))
        else:
            # We reuse the tokenizer hoping that will clear
            # ambiguities but this might be slow.
            spl = tokenizer(text)
    else:
        spl = list(text)

    spl = tuple(spl)
    if spl in existing:
        raise RuntimeError(
            f"The converter cannot guess how to split expression "
            f"{text!r} into tokens. This case happens when tokens have "
            f"spaces.")
    if (op.ngram_range[0] == 1 and
            (len(op.ngram_range) == 1 or op.ngram_range[1] > 1)):
        # All grams should be existing in the vocabulary.
        for g in spl:
            if g not in op.vocabulary_:
                raise RuntimeError(
                    "Unable to split n-grams '{}' into tokens {} "
                    "existing in the vocabulary. Token '{}' does not "
                    "exist in the vocabulary."
                    ".".format(text, spl, g))
    existing.add(spl)
    return spl


def convert_sklearn_text_vectorizer(scope: Scope, operator: Operator,
                                    container: ModelComponentContainer):
    """
    Converters for class
    `TfidfVectorizer <https://scikit-learn.org/stable/modules/generated/
    sklearn.feature_extraction.text.TfidfVectorizer.html>`_.
    The current implementation is a work in progress and the ONNX version
    does not produce the exact same results. The converter lets the user
    change some of its parameters.

    Additional options
    ------------------

    tokenexp: string
        The default will change to true in version 1.6.0.
        The tokenizer splits into words using this regular
        expression or the regular expression specified by
        *scikit-learn* is the value is an empty string.
        See also note below.
        Default value: None
    separators: list of separators
        These separators are used to split a string into words.
        Options *separators* is ignore if options *tokenexp* is not None.
        Default value: ``[' ', '[.]', '\\\\?', ',', ';', ':', '\\\\!']``.

    Example (from :ref:`l-example-tfidfvectorizer`):

    ::

        seps = {TfidfVectorizer: {"separators": [' ', '[.]', '\\\\?', ',', ';',
                                                 ':', '!', '\\\\(', '\\\\)',
                                                 '\\n', '\\\\"', "'", "-",
                                                 "\\\\[", "\\\\]", "@"]}}
        model_onnx = convert_sklearn(pipeline, "tfidf",
                                     initial_types=[("input", StringTensorType([None, 2]))],
                                     options=seps)

    The default regular expression of the tokenizer is ``(?u)\\\\b\\\\w\\\\w+\\\\b``
    (see `re <https://docs.python.org/3/library/re.html>`_).
    This expression may not supported by the library handling the backend.
    `onnxruntime <https://github.com/Microsoft/onnxruntime>`_ uses
    `re2 <https://github.com/google/re2>`_. You may need to switch
    to a custom tokenizer based on
    `python wrapper for re2 <https://pypi.org/project/re2/>`_
    or its sources `pyre2 <https://github.com/facebook/pyre2>`_
    (`syntax <https://github.com/google/re2/blob/master/doc/syntax.txt>`_).
    If the regular expression is not specified and if
    the instance of TfidfVectorizer is using the default
    pattern ``(?u)\\\\b\\\\w\\\\w+\\\\b``, it is replaced by
    ``[a-zA-Z0-9_]+``. Any other case has to be
    manually handled.

    Regular expression ``[^\\\\\\\\n]`` is used to split
    a sentance into character (and not works) if ``analyser=='char'``.
    The mode ``analyser=='char_wb'`` is not implemented.

    .. versionchanged:: 1.6
        Parameters have been renamed: *sep* into *separators*,
        *regex* into *tokenexp*.
    ````

    """  # noqa
    op = operator.raw_operator

    if (container.target_opset is not None and
            container.target_opset < 9):
        raise RuntimeError(
            "Converter for '{}' only works for opset >= 9."
            "".format(op.__class__.__name__))

    if op.analyzer == "char_wb":
        raise NotImplementedError(
            "CountVectorizer cannot be converted, "
            "only tokenizer='word' is fully supported. "
            "You may raise an issue at "
            "https://github.com/onnx/sklearn-onnx/issues.")
    if op.analyzer == "char":
        warnings.warn(
            "The conversion of CountVectorizer may not work. "
            "only tokenizer='word' is fully supported. "
            "You may raise an issue at "
            "https://github.com/onnx/sklearn-onnx/issues.",
            UserWarning)
    if op.strip_accents is not None:
        raise NotImplementedError(
            "CountVectorizer cannot be converted, "
            "only strip_accents=None is supported. "
            "You may raise an issue at "
            "https://github.com/onnx/sklearn-onnx/issues.")

    options = container.get_options(
        op, dict(separators="DEFAULT",
                 tokenexp=None,
                 nan=False,
                 keep_empty_string=False))
    if set(options) != {'separators', 'tokenexp', 'nan', 'keep_empty_string'}:
        raise RuntimeError("Unknown option {} for {}".format(
            set(options) - {'separators'}, type(op)))

    if op.analyzer == 'word':
        default_pattern = '(?u)\\b\\w\\w+\\b'
        if options['separators'] == "DEFAULT" and options['tokenexp'] is None:
            regex = op.token_pattern
            if regex == default_pattern:
                regex = '[a-zA-Z0-9_]+'
            default_separators = None
        elif options['tokenexp'] is not None:
            if options['tokenexp']:
                regex = options['tokenexp']
            else:
                regex = op.token_pattern
                if regex == default_pattern:
                    regex = '[a-zA-Z0-9_]+'
            default_separators = None
        else:
            regex = None
            default_separators = options['separators']
    else:
        if options['separators'] != 'DEFAULT':
            raise RuntimeError("Option separators has no effect "
                               "if analyser != 'word'.")
        regex = options['tokenexp'] if options['tokenexp'] else '.'
        default_separators = None

    if op.preprocessor is not None:
        raise NotImplementedError(
            "Custom preprocessor cannot be converted into ONNX. "
            "You may raise an issue at "
            "https://github.com/onnx/sklearn-onnx/issues.")
    if op.tokenizer is not None:
        raise NotImplementedError(
            "Custom tokenizer cannot be converted into ONNX. "
            "You may raise an issue at "
            "https://github.com/onnx/sklearn-onnx/issues.")
    if op.strip_accents is not None:
        raise NotImplementedError(
            "Operator StringNormalizer cannot remove accents. "
            "You may raise an issue at "
            "https://github.com/onnx/sklearn-onnx/issues.")

    if hasattr(op, "stop_words_"):
        stop_words = op.stop_words_ | (
            set(op.stop_words) if op.stop_words else set())
    else:
        stop_words = set()
    for w in stop_words:
        if not isinstance(w, str):
            raise TypeError(
                f"One stop word is not a string {w!r} "
                f"in stop_words={stop_words}.")

    if op.lowercase or stop_words:
        if len(operator.input_full_names) != 1:
            raise RuntimeError("Only one input is allowed, found {}.".format(
                operator.input_full_names))

        # StringNormalizer
        op_type = 'StringNormalizer'
        attrs = {'name': scope.get_unique_operator_name(op_type)}
        normalized = scope.get_unique_variable_name('normalized')
        if container.target_opset >= 10:
            attrs.update({
                'case_change_action': 'LOWER',
                'is_case_sensitive': not op.lowercase,
            })
            op_version = 10
            domain = ''
        else:
            attrs.update({
                'casechangeaction': 'LOWER',
                'is_case_sensitive': not op.lowercase,
            })
            op_version = 9
            domain = 'com.microsoft'
        opvs = 1 if domain == 'com.microsoft' else op_version
        if stop_words:
            attrs['stopwords'] = list(sorted(stop_words))

        if options['keep_empty_string']:
            del attrs['name']
            op_norm = OnnxStringNormalizer(
                'text_in', op_version=container.target_opset,
                output_names=['text_out'], **attrs)
            scan_body = op_norm.to_onnx(
                OrderedDict([('text_in', StringTensorType())]),
                outputs=[('text_out', StringTensorType())],
                target_opset=op_version)

            vector = scope.get_unique_variable_name('vector')
            apply_reshape(scope, operator.input_full_names[0],
                          vector, container,
                          desired_shape=(-1, 1))
            container.add_node('Scan', vector, normalized,
                               body=scan_body.graph, num_scan_inputs=1)
        else:
            flatten = scope.get_unique_variable_name('flattened')
            apply_reshape(scope, operator.input_full_names[0],
                          flatten, container,
                          desired_shape=(-1, ))
            container.add_node(op_type, flatten,
                               normalized, op_version=opvs,
                               op_domain=domain, **attrs)
    else:
        normalized = operator.input_full_names

    # Tokenizer
    padvalue = "#"
    while padvalue in op.vocabulary_:
        padvalue += "#"

    op_type = 'Tokenizer'
    attrs = {'name': scope.get_unique_operator_name(op_type)}
    attrs.update({
        'pad_value': padvalue,
        'mark': False,
        'mincharnum': 1,
    })
    if regex is None:
        attrs['separators'] = default_separators
    else:
        attrs['tokenexp'] = regex

    tokenized = scope.get_unique_variable_name('tokenized')
    container.add_node(op_type, normalized, tokenized,
                       op_domain='com.microsoft', **attrs)

    # Flatten
    # Tokenizer outputs shape {1, C} or {1, 1, C}.
    # Second shape is not allowed by TfIdfVectorizer.
    # We use Flatten which produces {1, C} in both cases.
    flatt_tokenized = scope.get_unique_variable_name('flattened')
    container.add_node("Flatten", tokenized, flatt_tokenized,
                       name=scope.get_unique_operator_name('Flatten'))
    tokenized = flatt_tokenized

    # Ngram - TfIdfVectorizer
    C = max(op.vocabulary_.values()) + 1
    words = [None for i in range(C)]
    weights = [0 for i in range(C)]
    for k, v in op.vocabulary_.items():
        words[v] = k
        weights[v] = 1.
    mode = 'TF'

    # Scikit-learn sorts n-grams by alphabetical order..
    # onnx assumes it is sorted by n.
    tokenizer = op.build_tokenizer()
    split_words = []
    existing = set()
    errors = []
    for w in words:
        if isinstance(w, tuple):
            # TraceableCountVectorizer, TraceableTfIdfVectorizer
            spl = list(w)
            w = ' '.join(w)
        else:
            # CountVectorizer, TfIdfVectorizer
            try:
                spl = _intelligent_split(w, op, tokenizer, existing)
            except RuntimeError as e:
                errors.append(e)
                continue
        split_words.append((spl, w))
    if len(errors) > 0:
        err = "\n".join(map(str, errors))
        raise RuntimeError(
            f"There were ambiguities between n-grams and tokens. "
            f"{len(errors)} errors occurred. You can fix it by using "
            f"class Traceable{op.__class__.__name__}.\n"
            f"You can learn more at https://github.com/scikit-learn/"
            f"scikit-learn/issues/13733.\n{err}")

    ng_split_words = sorted([(len(a[0]), a[0], i)
                            for i, a in enumerate(split_words)])
    key_indices = [a[2] for a in ng_split_words]
    ngcounts = [0 for i in range(op.ngram_range[0])]

    words = list(ng_split_words[0][1])
    for i in range(1, len(ng_split_words)):
        if ng_split_words[i - 1][0] != ng_split_words[i][0]:
            ngcounts.append(len(words))
        words.extend(ng_split_words[i][1])

    weights_ = [weights[a[2]] for a in ng_split_words]
    weights = list(weights_)
    for i, ind in enumerate(key_indices):
        weights[ind] = weights_[i]

    # Create the node.
    attrs = {'name': scope.get_unique_operator_name("TfIdfVectorizer")}
    attrs.update({
        'min_gram_length': op.ngram_range[0],
        'max_gram_length': op.ngram_range[1],
        'mode': mode,
        'max_skip_count': 0,
        'pool_strings': words,
        'ngram_indexes': key_indices,
        'ngram_counts': ngcounts,
        'weights': list(map(np.float32, weights)),
    })
    output = scope.get_unique_variable_name('output')

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    if proto_dtype == onnx_proto.TensorProto.DOUBLE:
        output_tf = scope.get_unique_variable_name('cast_result')
    else:
        output_tf = output

    if container.target_opset < 9:
        op_type = 'Ngram'
        container.add_node(op_type, tokenized, output_tf,
                           op_domain='com.microsoft', **attrs)
    else:
        op_type = 'TfIdfVectorizer'
        container.add_node(op_type, tokenized, output_tf, op_domain='',
                           op_version=9, **attrs)

    if proto_dtype == onnx_proto.TensorProto.DOUBLE:
        apply_cast(scope, output_tf, output,
                   container, to=proto_dtype)

    if op.binary:
        cast_result_name = scope.get_unique_variable_name('cast_result')
        output_name = scope.get_unique_variable_name('output_name')

        apply_cast(scope, output, cast_result_name, container,
                   to=onnx_proto.TensorProto.BOOL)
        apply_cast(scope, cast_result_name, output_name,
                   container, to=onnx_proto.TensorProto.FLOAT)
        output = output_name

    options = container.get_options(op, dict(nan=False))
    replace_by_nan = options.get('nan', False)
    if replace_by_nan:
        # This part replaces all null values by nan.
        cst_nan_name = scope.get_unique_variable_name('nan_name')
        container.add_initializer(cst_nan_name, proto_dtype, [1], [np.nan])
        cst_zero_name = scope.get_unique_variable_name('zero_name')
        container.add_initializer(cst_zero_name, proto_dtype, [1], [0])

        mask_name = scope.get_unique_variable_name('mask_name')
        container.add_node('Equal', [output, cst_zero_name],
                           mask_name,
                           name=scope.get_unique_operator_name('Equal'))

        where_name = scope.get_unique_variable_name('where_name')
        container.add_node('Where', [mask_name, cst_nan_name, output],
                           where_name,
                           name=scope.get_unique_operator_name('Where'))
        output = where_name

    apply_identity(scope, output, operator.output_full_names, container)


register_converter('SklearnCountVectorizer', convert_sklearn_text_vectorizer,
                   options={'tokenexp': None, 'separators': None,
                            'nan': [True, False],
                            'keep_empty_string': [True, False]})



# SPDX-License-Identifier: Apache-2.0


import re
import warnings
import pprint
from logging import getLogger
from collections import OrderedDict
import numpy as np
from onnx import onnx_pb as onnx_proto
from onnxconverter_common.data_types import (  # noqa
    DataType, TensorType,
    FloatType, Int64Type, StringType,
    DictionaryType, FloatTensorType,  # noqa
    Int64TensorType, SequenceType,  # noqa
    StringTensorType, DoubleTensorType,
    Int32TensorType, BooleanTensorType,
    DoubleTensorType)
try:
    from onnxconverter_common.data_types import (
        Int8TensorType, UInt8TensorType)
except ImportError:
    Int8TensorType = None
    UInt8TensorType = None
from ..proto import (
    get_opset_number_from_onnx,
    get_latest_tested_opset_version
)
from ..proto.onnx_helper_modified import (
    make_graph, make_model, make_tensor_value_info
)
from . import _registration
from . import utils
from .exceptions import MissingShapeCalculator, MissingConverter
from ._container import ModelComponentContainer, _build_options
from .onnx_optimisation_identity import onnx_remove_node_identity

type_fct = type


def _default_OPSET_TO_IR_VERSION():
    return {
        1: 3, 2: 3, 3: 3, 4: 3, 5: 3, 6: 3,
        7: 3, 8: 4, 9: 4, 10: 5, 11: 6, 12: 7,
        13: 7, 14: 7, 15: 8, 16: 8, 17: 8, 18: 8
    }


try:
    from onnxconverter_common.topology import OPSET_TO_IR_VERSION
    assert OPSET_TO_IR_VERSION[18] is not None
except (ImportError, KeyError):
    OPSET_TO_IR_VERSION = _default_OPSET_TO_IR_VERSION()

OPSET_ML_TO_OPSET = {1: 11, 2: 15, 3: 18}

logger = getLogger('skl2onnx')


def get_default_opset_for_domain(domain):
    """
    Returns the associated for a domain given the main opset.
    """
    from .. import __max_supported_opset__ as main_opset
    if domain == '':
        return main_opset
    if domain == 'ai.onnx.ml':
        if main_opset >= 16:
            return 3
        if main_opset < 6:
            return 1
        return 2
    if domain == 'ai.onnx.training':
        return 1
    return None


class Variable:
    """
    Defines a variable which holds any data defined
    from *ONNX* types.
    """
    _UNIQUE_NUMBER_ = 0

    def __init__(self, raw_name, onnx_name, scope, type=None):
        """
        :param raw_name: A string indicating the variable's name in the
                         original model. Usually, it's the seed string
                         used to created its ONNX name (i.e., the
                         field *onnx_name* below).
        :param onnx_name: A string indicating the variable's name in
                          the converted model
        :param scope: A string. It's the name of the scope where this
                      variable is declared
        :param type: A type object defined in .common.data_types.py;
                     e.g., FloatTensorType
        """
        if not isinstance(raw_name, str):
            raise TypeError(
                "raw_name must be a string not '%s'." % raw_name.__class__)
        if type is not None and not hasattr(type, 'shape'):
            raise TypeError(
                "Unexpected type for variable raw_name=%r, type=%r." % (
                    raw_name, type))
        if not isinstance(onnx_name, str) or '(' in onnx_name:
            if onnx_name.startswith('u(') and onnx_name[-1] == ')':
                onnx_name0 = onnx_name
                if scope is None:
                    onnx_name = "UU%03dUU" % Variable._UNIQUE_NUMBER_
                    Variable._UNIQUE_NUMBER_ += 1
                else:
                    onnx_name = scope.get_unique_variable_name("U")
                logger.debug(
                    '[Var] rename raw_name=%r, onnx_name=%r into %r',
                    raw_name, onnx_name0, onnx_name)
            else:
                raise TypeError(
                    "onnx_name must be a string not %r." % onnx_name)

        if type is not None:
            shape = type.shape
            if shape is not None:
                not_none = [v for v in shape if v is not None]
                if len(not_none) and min(not_none) == 0:
                    raise RuntimeError(
                        "A variable cannot be empty, raw_name=%r, "
                        "onnx_name=%r, shape=%r, type=%r." % (
                            raw_name, onnx_name, shape, type))

        self._raw_name = raw_name
        self._onnx_name = onnx_name
        self._scope = scope
        self._type = type
        self._parent = None

        # The following fields are bool variables used in parsing and
        # compiling stages
        self._is_fed = None
        self._is_root = None
        self._is_leaf = None
        if self.type is not None and not isinstance(self.type, DataType):
            raise TypeError(
                "shape must be a DataType not {}.".format(self.type))
        if isinstance(self.type, TensorType):
            shape = self.type.shape
            if not isinstance(shape, (list, tuple)):
                try:
                    shape = list(shape)
                except TypeError:
                    raise TypeError("shape must be a tuple or a list not "
                                    "{}.".format(type_fct(shape)))
            for dim in shape:
                if dim is None:
                    continue
                if not isinstance(dim, (int, np.int32, np.int64, np.intc)):
                    raise TypeError(
                        "shape must contains integers not %r (type=%r)."
                        "" % (dim, dim.__class__))
        logger.debug('[Var] +%s', self)

        # links to operators using those variables
        self.operators_outputs_ = []
        self.operators_inputs_ = []
        self._check()

    def _check(self):
        if self.type is not None and self.type.shape is not None:
            for k in self.type.shape:
                if k is None:
                    continue
                if not isinstance(k, (int, np.integer)):
                    raise ValueError(
                        "Unexpected type %r for shape %r."
                        "" % (type(k), self))

    @property
    def raw_name(self):
        return self._raw_name

    @property
    def onnx_name(self):
        return self._onnx_name

    @property
    def scope(self):
        return self._scope

    @property
    def type(self):
        return self._type

    @property
    def is_fed(self):
        return self._is_fed

    @property
    def is_root(self):
        return self._is_root

    @property
    def is_leaf(self):
        return self._is_leaf

    def init_status(self, is_fed=None, is_root=None, is_leaf=None):
        if is_fed is not None and is_fed != self.is_fed:
            logger.debug(
                '[Var] update is_fed=%r for %r, parent=%r',
                is_fed, self, self._parent)
            self._is_fed = is_fed
        if is_root is not None and is_root != self.is_root:
            logger.debug('[Var] update is_root=%r for %r', is_root, self)
            self._is_root = is_root
        if is_leaf is not None and is_leaf != self.is_leaf:
            logger.debug('[Var] update is_leaf=%r for %r', is_leaf, self)
            self._is_leaf = is_leaf

    def __setattr__(self, name, value):
        if name == "type":
            self.set_type(value)
        elif name == "onnx_name":
            raise AttributeError("You must use method set_onnx_name.")
        elif name in {"is_fed", "is_root", "is_leaf"}:
            raise AttributeError("You must use method init_status.")
        elif name in {'scope', 'raw_name'}:
            raise AttributeError("scope or raw_name cannot be changed.")
        self.__dict__[name] = value

    def set_type(self, new_type):
        if (new_type is None or isinstance(new_type, (str, Variable)) or
                not hasattr(new_type, 'shape')):
            raise TypeError(
                "Unexpected new type for variable %r, new_type=%r." % (
                    self, new_type))
        logger.debug('[Var] update type for %r', self)
        self._type = new_type
        self._check()

    def set_onnx_name(self, onnx_name):
        if onnx_name != self._onnx_name:
            logger.debug(
                '[Var] update onnx_name, from %r to %r in %r',
                self.onnx_name, onnx_name, self)
            if self.scope is not None and not isinstance(self.scope, str):
                self.scope.rename_onnx_name(self._onnx_name, onnx_name)
            self._onnx_name = onnx_name

    def set_parent(self, operator):
        if self._parent is not None:
            raise RuntimeError(
                "This variable is already the output of operator %r. "
                "It cannot be the output of %r." % (self._parent, operator))
        logger.debug(
            '[Var] set parent for %r, parent=%r', self, operator)
        self._parent = operator

    def get_first_dimension(self):
        """
        Returns the first dimension (batch dimension) or
        None if not specified (shape is empty).
        """
        if (self.type is None or self.type.shape is None or
                len(self.type.shape) == 0):
            return None
        return self.type.shape[0]

    def get_second_dimension(self):
        if (self.type is None or self.type.shape is None or
                len(self.type.shape) < 2):
            return None
        return self.type.shape[1]

    @property
    def full_name(self):
        """
        Return a globally unique variable ID
        """
        return self.onnx_name

    def __repr__(self):
        return ("Variable('{0}', '{1}', type={2})".format(
                self.raw_name, self.onnx_name, self.type))

    @staticmethod
    def from_pb(obj):
        """
        Creates a data type from a protobuf object.
        """
        def get_dim(d):
            r = d.dim_value
            if "dim_param" in str(d):
                return None
            if r == 0:
                # dim_value is 0 when it is 0 or undefined
                return 0 if "0" in str(d) else None
            return r

        def get_shape(tt):
            return [get_dim(tt.shape.dim[i])
                    for i in range(len(tt.shape.dim))]

        if hasattr(obj, 'extend'):
            return [Variable.from_pb(o) for o in obj]

        name = obj.name
        if obj.type.tensor_type:
            tt = obj.type.tensor_type
            elem = tt.elem_type
            shape = get_shape(tt)
            if elem == onnx_proto.TensorProto.FLOAT:
                ty = FloatTensorType(shape)
            elif elem == onnx_proto.TensorProto.BOOL:
                ty = BooleanTensorType(shape)
            elif elem == onnx_proto.TensorProto.DOUBLE:
                ty = DoubleTensorType(shape)
            elif elem == onnx_proto.TensorProto.STRING:
                ty = StringTensorType(shape)
            elif elem == onnx_proto.TensorProto.INT64:
                ty = Int64TensorType(shape)
            elif elem == onnx_proto.TensorProto.INT32:
                ty = Int32TensorType(shape)
            elif (UInt8TensorType is not None and
                    elem == onnx_proto.TensorProto.UINT8):
                ty = UInt8TensorType(shape)
            elif (Int8TensorType is not None and
                    elem == onnx_proto.TensorProto.INT8):
                ty = Int8TensorType(shape)
            elif elem == 0:
                ty = FloatTensorType(shape)
            else:
                raise NotImplementedError(
                    "Unsupported type '{}' (elem_type={}).".format(
                        type(obj.type.tensor_type), elem))
        else:
            raise NotImplementedError("Unsupported type '{}' as "
                                      "a string ({}).".format(
                                          type(obj), obj))

        return Variable(name, name, None, ty)

    def __iter__(self):
        "Enables expression such as `a,b = self`."
        yield self.onnx_name
        yield self.type

    def __getitem__(self, index):
        if index == 0:
            return self.onnx_name
        if index == 1:
            return self.type
        raise IndexError("Unreachable element at index %d." % index)

    def add_operator(self, op, in_or_out):
        "Add a link to an operator, True for output, False for input."
        if in_or_out:
            self.operators_outputs_.append(op)
        else:
            self.operators_inputs_.append(op)

    def check_compatible_type(self, other_type):

        def empty_shape(shape):
            return shape is None or len(shape) == 0

        if self.type is None:
            if other_type is None:
                return
        elif other_type is not None:
            if isinstance(self.type, type(other_type)):
                if self.type.shape == other_type.shape:
                    return
                if empty_shape(other_type.shape):
                    return
        raise TypeError(
            "Incompatible type for variable %r and type %r." % (
                self, other_type))


class VariableStr(Variable):
    """
    Defines a variable a string. This should be avoided.
    """

    def __init__(self, name, scope=None, type=None):
        Variable.__init__(self, name, name, scope=scope, type=type)

    @property
    def raw_name(self):
        return self._raw_name

    @property
    def onnx_name(self):
        if self._onnx_name.startswith("u("):
            raise RuntimeError(
                "Variable should be renamed as onnx_name=%r."
                "" % self._onnx_name)
        return self._onnx_name


class Operator:
    """
    Defines an operator available in *ONNX*.
    """
    class OperatorList(list):
        def __init__(self, parent, kind):
            super(Operator.OperatorList, self).__init__()
            self.parent = parent
            self.kind = kind

        def __eq__(self, second):
            raise NotImplementedError(
                "Operator equal not implemented and not needed.")

        def append(self, v):
            if not isinstance(v, Variable):
                raise TypeError(
                    "Input and output must be of type Variable not %r."
                    "" % type(v))
            if self.kind == 'Out':
                v.set_parent(self.parent)
            super(Operator.OperatorList, self).append(v)
            logger.debug("[Op] add %s %r to %r", self.kind, v, self.parent)
            if self.kind == 'In':
                v.add_operator(self.parent, False)
            elif self.kind == "Out":
                v.add_operator(self.parent, True)
            else:
                raise RuntimeError(
                    "Unexpected value for kind=%r." % self.kind)

        def extend(self, vs):
            for v in vs:
                self.append(v)

        def __getitem__(self, i):
            v = list.__getitem__(self, i)
            if isinstance(i, int) and not isinstance(v, Variable):
                raise TypeError("Element %d must be a Variable not %r." % (
                    i, type(v)))
            return v

        def __setitem__(self, i, v):
            raise LookupError(
                "Setter should not be used to modify an element.")

        def set_element(self, i, v):
            "Updates element i."
            if not isinstance(v, Variable):
                raise TypeError(
                    "Value v must be a Variable not %r." % type(v))
            logger.debug(
                "[Op] %s-change element %d from %r to %r in %r",
                self.kind, i, self[i], v, self.parent)
            list.__setitem__(self, i, v)

        def to_string(self):
            names = []
            for o in self:
                if hasattr(o, 'onnx_name'):
                    names.append(o.onnx_name)
                else:
                    names.append('"%s"' % str(o))
            return ",".join(names)

    def __init__(self, onnx_name, scope, type, raw_operator,
                 target_opset, scope_inst):
        """
        :param onnx_name: A unique ID, which is a string
        :param scope: The name of the scope where this operator is
                      declared. It's a string.
        :param type: A object which uniquely characterizes the type of
                     this operator. For example, it can be a string,
                     pooling, if this operator is associated with a
                     CoreML pooling layer.
        :param raw_operator: The original operator which defines this operator;
                             for example, a scikit-learn Imputer and
                             a CoreML Normalizer.
        :param target_opset: The target opset number for the converted model.
        :param scope_inst: :class:`Scope` instance the operator belongs to
        """
        if isinstance(raw_operator, str):
            raise RuntimeError("Parameter raw_operator must be an object not "
                               "a string '{0}'.".format(raw_operator))
        # operator name in the converted model, if raw_operator
        # is not None, output_shapes can be guessed
        # from the raw model. Otherwise, it can be guessed
        # from the input shapes.
        self.onnx_name = onnx_name
        self.scope = scope
        self.type = type
        self.raw_operator = raw_operator
        self.inputs = Operator.OperatorList(self, 'In')
        self.outputs = Operator.OperatorList(self, 'Out')
        self._is_evaluated = None
        self.target_opset = target_opset
        self.scope_inst = scope_inst
        logger.debug('[Op] +%r', self)

    def new_raw_operator(self, raw_operator, alias):
        """
        Returns a shallow copy of this operator,
        changes the raw_operator but keeps the same inputs
        and outputs.
        """
        op = Operator(self.onnx_name, self.scope, alias, raw_operator,
                      self.target_opset, self.scope_inst)
        op.inputs = self.inputs
        op.outputs = self.outputs
        return op

    def __repr__(self):
        try:
            textop = repr(self.raw_operator)
        except AttributeError:
            textop = "MISSING OP"
        except KeyError:
            # The line above fails for python 3.7
            textop = type(self.raw_operator)
        if isinstance(textop, str) and "\n" in textop:
            textop = textop.replace('\n', '').replace(' ', '')
        return ("Operator(type='{0}', onnx_name='{1}', inputs='{2}', "
                "outputs='{3}', raw_operator={4})".format(
                    self.type, self.onnx_name,
                    self.inputs.to_string(),
                    self.outputs.to_string(),
                    textop))

    def __setattr__(self, name, value):
        if name in ('inputs', 'outputs'):
            if (isinstance(value, list) and
                    not isinstance(value, Operator.OperatorList)):
                if name == 'inputs':
                    self.inputs = Operator.OperatorList(self, 'In')
                    self.inputs.extend(value)
                    return
                if name == 'outputs':
                    self.outputs = Operator.OperatorList(self, 'Out')
                    self.outputs.extend(value)
                    return
            if not isinstance(value, Operator.OperatorList):
                raise TypeError(
                    "inputs or outputs must be of type Operator.OperatorList.")
            ioo = name == 'outputs'
            for v in value:
                v.add_operator(self, ioo)
        self.__dict__[name] = value

    @property
    def is_evaluated(self):
        return self._is_evaluated

    def init_status(self, is_evaluated=None):
        if is_evaluated is not None and is_evaluated != self.is_evaluated:
            logger.debug(
                '[Op] update is_evaluated=%r for %r',
                is_evaluated, self)
            self._is_evaluated = is_evaluated

    @property
    def full_name(self):
        """
        Return a globally unique operator ID
        """
        return self.onnx_name

    @property
    def input_full_names(self):
        """
        Return all input variables' names
        """
        return [variable.full_name for variable in self.inputs]

    @property
    def output_full_names(self):
        """
        Return all output variables' names
        """
        return [variable.full_name for variable in self.outputs]

    @property
    def original_operator(self):
        """
        Return the original operator/layer
        """
        return self.raw_operator

    def infer_types(self):
        # Invoke a core inference function
        if self.type is None:
            raise MissingShapeCalculator(
                "Unable to find a shape calculator for type '{}'.".format(
                    type(self.raw_operator)))
        try:
            shape_calc = _registration.get_shape_calculator(self.type)
        except ValueError:
            raise MissingShapeCalculator(
                "Unable to find a shape calculator for alias '{}' "
                "and type '{}'.".format(self.type, type(self.raw_operator)))
        if shape_calc is None:
            raise MissingShapeCalculator(
                "Unexpected shape calculator for alias '{}' "
                "and type '{}'.".format(self.type, type(self.raw_operator)))
        logger.debug(
            "[Shape-a] %r fed %r - %r", self,
            "".join(str(i.is_fed) for i in self.inputs),
            "".join(str(i.is_fed) for i in self.outputs))
        shape_calc(self)
        logger.debug(
            "[Shape-b] %r inputs=%r - outputs=%r",
            self, self.inputs, self.outputs)


class Scope:
    """
    Every node of an *ONNX* graph must be unique. This class holds the list
    of existing name for every node already defined in graph. It also
    provides functions to create a unique unused name.
    """

    def __init__(self, name, target_opset=None,
                 custom_shape_calculators=None, options=None,
                 registered_models=None, naming=None):
        """
        :param name: A string, the unique ID of this scope in a
                     Topology object
        :param target_opset: The target opset number for the converted
                             model.
        :param custom_conversion_functions: a dictionary for specifying
                                the user customized conversion function
        :param custom_shape_calculators: a dictionary for specifying
                                the user customized shape calculator
        :param options: see :ref:`l-conv-options`
        :param naming: the user may want to change the way intermediate
            are named, this parameter can be a string (a prefix) or a
            function, which signature is the following:
            `get_name(name, existing_names)`, the library will then
            check this name is unique and modify it if not
        :param registered_models: registered models

        .. versionchanged:: 1.10.0
            Parameter *naming* was added.
        """
        self.name = name
        self.onnx_variable_names = set()
        self.onnx_operator_names = set()
        self.target_opset = target_opset
        self.custom_shape_calculators = custom_shape_calculators

        # An one-to-many map from raw variable name to ONNX variable
        # names. It looks like
        # (key, value) = (raw_name, [onnx_name, onnx_name1, onnx_name2, ..., onnx_nameN]) # noqa
        # The last name may hide all other names in this scope.
        self.variable_name_mapping = {}

        # A map of local variables defined in this scope.
        # (key, value) = (onnx_name, variable)
        self.variables = OrderedDict()
        self.input_variables = []
        self.output_variables = []

        # A map of local operators defined in this scope.
        # (key, value) = (onnx_name, operator)
        self.operators = {}

        # Additional options given to converters.
        self.options = options

        # Registered models
        self.registered_models = registered_models
        self.naming = naming

        if naming is None:
            self._naming = Topology._generate_unique_name
        elif isinstance(naming, str):
            self._naming = (
                lambda seed, names: Topology._generate_unique_name(
                    self.naming + seed, names))
        elif callable(self.naming):
            self._naming = (
                lambda seed, names: Topology._generate_unique_name(
                    self.naming(seed, names), names))
        else:
            raise TypeError(
                "Unexpected type for parameter naming: %r." % type(naming))

    def get(self, var_name, default_value):
        "Returns variable with 'name' or default value is not found."
        return self.variables.get(var_name, default_value)

    def has_variable_name(self, name):
        """
        Tells if a variable is already registered.
        """
        return name in self.onnx_variable_names

    def get_shape_calculator(self, model_type):
        """
        Returns the shape calculator for the given model type.

        :param model_type: model type such as *LogisticRegression*
        :return: alias or None if not found
        """
        return self.custom_shape_calculators.get(model_type, None)

    def get_unique_variable_name(self, seed, rename=True):
        """
        Creates a unique variable ID based on the given seed.
        """
        if not isinstance(seed, str):
            raise TypeError("Parameter seed must be a string not {}."
                            "".format(type(seed)))
        if rename:
            name = self._naming(seed, self.onnx_variable_names)
        else:
            name = Topology._generate_unique_name(
                seed, self.onnx_variable_names)
        return name

    def get_unique_operator_name(self, seed):
        """
        Creates a unique operator ID based on the given seed.
        """
        return self._naming(seed, self.onnx_operator_names)

    def declare_local_variable(self, raw_name, type=None, prepend=False,
                               missing_type=False, rename=True):
        """
        This function may create a new variable in this scope. If
        *raw_name* has been used to create other variables, the new
        variable will hide all other variables created using *raw_name*.
        """
        if type is None and not missing_type:
            raise RuntimeError(
                "Unknown type for %r (type=%r)." % (raw_name, type))
        # Get unique ID for the new variable
        onnx_name = self.get_unique_variable_name(raw_name, rename=rename)

        # Create the variable
        variable = Variable(raw_name, onnx_name, self.name, type)
        self.register_variable(variable, prepend=prepend)
        return variable

    def register_variable(self, var, prepend=False):
        "Adds a variable to the scope."
        if var.onnx_name in self.variables:
            raise RuntimeError(
                "Variable %r already registered (other=%r)." % (
                    var, self.variables[var.onnx_name]))

        if var.raw_name in self.variable_name_mapping:
            # Hide existing variables with the same raw_name
            if not prepend:
                self.variable_name_mapping[var.raw_name].append(var.onnx_name)
            else:
                self.variable_name_mapping[var.raw_name].insert(
                    0, var.onnx_name)
        else:
            self.variable_name_mapping[var.raw_name] = [var.onnx_name]

        self.variables[var.onnx_name] = var

    def declare_existing_subgraph_name(self, graph_proto):
        """
        Declare all name from a subgraph in order to avoid being picked twice.
        """
        output_name = {o.name for o in graph_proto.output}
        for node in graph_proto.node:
            for name in node.output:
                if name in output_name:
                    continue
                if self.has_variable_name(name):
                    raise NameError(
                        "Result name %r is already taken (outputs=%r) "
                        "(node=%r)." % (
                            name, output_name, node))
                self.onnx_variable_names.add(name)
            if node.name in self.onnx_operator_names:
                raise NameError(
                    "Operator name %r is already taken "
                    "(node=%r)." % (
                        node.name, node))
            self.onnx_operator_names.add(node.name)

    def rename_onnx_name(self, old_name, new_name):
        if new_name in self.variables:
            raise RuntimeError(
                "Name %r already in variables (%r)." % (
                    new_name, self.variables[new_name]))
        if old_name not in self.variables:
            raise RuntimeError(
                "Unable to find name %r in variables." % old_name)
        logger.debug(
            '[Scope] update onnx_name, from %r to %r',
            old_name, new_name)
        self.variables[new_name] = self.variables[old_name]
        del self.variables[old_name]

    def declare_local_input(self, raw_name, type=None, prepend=False,
                            rename=True):
        """
        Calls `declare_local_variable`. Registers this variable
        as an input.
        """
        var = self.declare_local_variable(
            raw_name, type=type, prepend=prepend, rename=rename)
        self.input_variables.append(var)
        return var

    def declare_local_output(self, raw_name, type=None, prepend=False,
                             missing_type=False):
        """
        Calls `declare_local_variable`. Registers this variable
        as an output.
        """
        var = self.declare_local_variable(
            raw_name, type=type, prepend=prepend,
            missing_type=missing_type)
        self.output_variables.append(var)
        return var

    def declare_local_operator(self, type, raw_model=None):
        """
        This function is used to declare new local operator.
        """
        onnx_name = self.get_unique_operator_name(str(type))
        operator = Operator(onnx_name, self.name, type, raw_model,
                            self.target_opset, scope_inst=self)
        self.operators[onnx_name] = operator
        return operator

    def _get_allowed_options(self, model, fail=True):
        if self.registered_models is not None:
            if type(model) not in self.registered_models['aliases']:
                if fail:
                    raise NotImplementedError(
                        "No registered models, no known allowed options "
                        "for model '{}'.".format(model.__class__.__name__))
                return {}
            alias = self.registered_models['aliases'][type(model)]
            conv = self.registered_models['conv'][alias]
            allowed = conv.get_allowed_options()
            return allowed
        raise NotImplementedError(
            "No registered models, no known allowed options "
            "for model '{}'.".format(model.__class__.__name__))

    def add_options(self, model_id, options):
        """
        Adds an option, for example,
        ``add_options(id(clr), {'raw_scores': True})``
        tells the converter associated to ``clr`` to
        use raw score instead of probabilities.

        :param model_id: class or ``id(instance)``
        :param options: dictionary with the new values
        """
        if options is None:
            return
        if self.options is None:
            self.options = {}
        if model_id not in self.options:
            self.options[model_id] = None
        if self.options[model_id] is None:
            self.options[model_id] = {}
        self.options[model_id].update(options)

    def get_options(self, model, default_values=None, fail=True):
        """
        Returns additional options for a model.
        It first looks by class then by id (``id(model)``).
        :param model: model being converted
        :param default_values: default options (it is modified by
                               the function)
        :param fail: fails if option it not found
        :return: dictionary
        """
        return _build_options(
            model, self.options, default_values,
            self._get_allowed_options(model, fail=fail),
            fail=fail)

    def replace_raw_operator(self, op1, op2, alias):
        """
        Replaces every raw operator op1 by op2.
        The function uses `id()` to detect op1.
        """
        for v in self.operators.values():
            if id(v.raw_operator) == id(op1):
                logger.debug(
                    '[Scope] replace %d by %d in %r.',
                    id(v.raw_operator), id(op1), v)
                v.raw_operator = op2
                v.type = alias


class Topology:
    """
    Holds instances on :class:`Scope <skl2onnx.common._topology.Scope>` and
    :class:`SklearnModelContainer
    <skl2onnx.common._container.SklearnModelContainer>`.
    These are filled by the converters while a pipeline is being converted.
    """

    def __init__(self, model, default_batch_size=1, initial_types=None,
                 target_opset=None, custom_conversion_functions=None,
                 custom_shape_calculators=None, registered_models=None):
        """
        Initializes a *Topology* object, which is an intermediate
        representation of a computational graph.

        :param model: RawModelContainer object or one of its derived
                      classes. It contains the original model.
        :param default_batch_size: batch_size prepend to scalar and
                                   array types from CoreML. It's usually
                                   1 or None.
        :param initial_types: A list providing some types for some
                              root variables.
        Each element is a tuple of a variable name and a type defined
        in *data_types.py*.
        :param custom_conversion_functions: a dictionary for specifying
                                the user customized conversion function
        :param custom_shape_calculators: a dictionary for specifying the
                                        user customized shape calculator
        :param registered_models: registered models
        """
        self.scopes = []
        self.raw_model = model
        self.scope_names = set()
        self.initial_types = initial_types if initial_types else list()
        self.default_batch_size = default_batch_size
        self.target_opset = target_opset
        self.custom_conversion_functions = (
            custom_conversion_functions if custom_conversion_functions else {})
        self.custom_shape_calculators = (
            custom_shape_calculators if custom_shape_calculators else {})

        for k in self.custom_conversion_functions:
            if not callable(k):
                raise TypeError("Keys in custom_conversion_functions must be "
                                "types not strings.")
        for k in self.custom_shape_calculators:
            if not callable(k):
                raise TypeError("Keys in custom_shape_calculators must be "
                                "types not strings.")

        # A map of local overwritten model aliases.
        self.model_aliases = {}
        all_model_types = (set(self.custom_conversion_functions)
                           | set(self.custom_shape_calculators))
        for mtype in all_model_types:
            alias = "{}_{}".format(mtype.__name__, id(self))
            self.model_aliases[mtype] = alias

        # Registered models
        if registered_models is None:
            raise AssertionError()
        self.registered_models = registered_models

    @property
    def scope(self):
        if len(self.scopes) != 1:
            raise RuntimeError(
                "Only one scope is allowed not %d." % len(self.scopes))
        return self.scopes[0]

    @staticmethod
    def _generate_unique_name(seed, existing_names):
        """
        Produce an unique string based on the seed
        :param seed: a string
        :param existing_names: a set containing strings which cannot be
                               produced
        :return: a string similar to the seed
        """
        if seed == '':
            raise ValueError('Name seed must be a non-empty string.')

        # Make the seed meet C-style naming convention
        # Only alphabets and numbers are allowed
        seed = re.sub('[^\\w+]', '_', seed)
        # The first symbol cannot be a number
        if re.match('^[0-9]', seed):
            seed = '_' + seed

        # If seed has never been seen, we return it as it is. Otherwise,
        # we will append an number to make it unique.
        if seed not in existing_names:
            existing_names.add(seed)
            return seed
        else:
            i = 1
            while seed + str(i) in existing_names:
                i += 1
            new_name = seed + str(i)
            existing_names.add(new_name)
            return new_name

    def get_unique_scope_name(self, seed):
        return Topology._generate_unique_name(seed, self.scope_names)

    def declare_scope(self, seed, parent_scopes=None, options=None,
                      naming=None):
        """
        Creates a new :class:`Scope <skl2onnx.common._topology.Scope>`
        and appends it to the list of existing scopes.
        """
        if len(self.scopes) != 0:
            raise RuntimeError(
                "Only one scope can be created.")
        scope = Scope(
            self.get_unique_scope_name(seed), target_opset=self.target_opset,
            custom_shape_calculators=self.custom_shape_calculators,
            options=options, registered_models=self.registered_models,
            naming=naming)

        # Declare input variables.
        # They should be the inputs of the scikit-learn
        # model you want to convert into ONNX.
        for var_name, initial_type in self.initial_types:
            scope.declare_local_input(var_name, initial_type, rename=False)
        self.scopes.append(scope)
        return scope

    def unordered_operator_iterator(self):
        for scope in self.scopes:
            for operator in scope.operators.values():
                yield operator

    def unordered_variable_iterator(self):
        for scope in self.scopes:
            for variable in scope.variables.values():
                yield variable

    def call_converter(self, operator, container, verbose=0):
        "Calls converter for operator *operator*."
        mtype = type(operator.raw_operator)
        if mtype in self.custom_conversion_functions:
            conv = self.custom_conversion_functions[mtype]
        elif operator.type in self.custom_conversion_functions:
            conv = self.custom_conversion_functions[operator.type]
        elif hasattr(operator.raw_operator, "onnx_converter"):
            conv = operator.raw_operator.onnx_converter()
        else:
            # Convert the selected operator into some ONNX objects and
            # save them into the container
            try:
                conv = _registration.get_converter(operator.type)
            except ValueError:
                raise MissingConverter(
                    "Unable to find converter for alias '{}' type "
                    "'{}'. You may raise an issue at "
                    "https://github.com/onnx/sklearn-onnx/issues."
                    "".format(operator.type,
                              type(getattr(operator, 'raw_model', None))))

        container.validate_options(operator)
        if verbose > 0:
            print("[call_converter] call converter for %r." % operator.type)
        logger.debug(
            "[Conv] call %r fed %r - %r", operator,
            "".join(str(i.is_fed) for i in operator.inputs),
            "".join(str(i.is_fed) for i in operator.outputs))
        conv(self.scopes[0], operator, container)
        logger.debug("[Conv] end - %r", operator)

    def call_shape_calculator(self, operator):
        "Calls shape_calculator for operator *operator*."
        mtype = type(operator.raw_operator)
        if mtype in self.custom_shape_calculators:
            # overwritten operator.
            source = 'custom'
            shape_calc = self.custom_shape_calculators[mtype]
        elif operator.type in self.custom_shape_calculators:
            source = 'custom'
            shape_calc = self.custom_shape_calculators[operator.type]
        elif hasattr(operator.raw_operator, "onnx_shape_calculator"):
            source = 'onnx_shape_calculator'
            shape_calc = operator.raw_operator.onnx_shape_calculator()
        else:
            source = ""
            shape_calc = None

        if shape_calc is not None:
            logger.debug(
                "[Shape1] %r fed %r - %r (source=%r)", operator,
                ",".join(str(i.is_fed) for i in operator.inputs),
                ",".join(str(i.is_fed) for i in operator.outputs),
                source)
            shape_calc(operator)
        else:
            logger.debug('[Shape2] call infer_types for %r', operator)
            operator.infer_types()

    def _initialize_graph_status_for_traversing(self):
        """
        Initialize the status of all variables and operators before
        traversing the graph. Only used by convert_operators.
        """
        if len(self.scopes) != 1:
            raise RuntimeError(
                "Only one scope is allowed not %d." % len(self.scopes))
        input_names = set(v.onnx_name for v in self.scopes[0].input_variables)
        if len(input_names) == 0:
            raise RuntimeError("No detected inputs.")
        for variable in self.unordered_variable_iterator():
            is_input = variable.onnx_name in input_names
            variable.init_status(is_fed=is_input)

        for operator in self.unordered_operator_iterator():
            operator.init_status(is_evaluated=False)

    def _propagate_status(self, operator, container, fed_variables,
                          verbose=0):
        """
        Propagates status *is_fed* based on output variable
        and node added in the container.
        """
        if verbose > 1:
            print("[_propagate_status] after op=%r" % operator)
        vars = {}
        for node in container.nodes:
            for i in node.input:
                if i not in vars:
                    vars[i] = []
                vars[i].append(node)

        if verbose > 1:
            print("[_propagate_status] newly fed=%r" % list(
                v.onnx_name for v in operator.outputs if v.is_fed))
        stack = list(fed_variables)
        scope = self.scopes[0]
        while len(stack) > 0:
            nodes = {}
            for name in stack:
                if name not in vars:
                    continue
                for n in vars[name]:
                    nodes[id(n)] = n
            stack = []
            for node in nodes.values():
                if all(fed_variables.get(n, False) for n in node.input):
                    for o in node.output:
                        if o not in fed_variables:
                            if verbose > 1:
                                print("[_propagate_status] add=%r" % o)
                            fed_variables[o] = o
                            stack.append(o)
                            if o in scope.variables:
                                var = scope.variables[o]
                                var.init_status(is_fed=True)
                                if verbose > 1:
                                    print("[_propagate_status] fed=%r" % var)

    def convert_operators(self, container=None, verbose=0):
        """
        Calls all converters and shape_calculator for existing
        operators. It also processes new operators created by
        converters.
        """
        def _check_operator_(operator):
            if not isinstance(operator.inputs, Operator.OperatorList):
                raise TypeError(
                    "operator.inputs must be a Operator.OperatorList "
                    "not %r." % type(operator.inputs))
            if not isinstance(operator.outputs, Operator.OperatorList):
                raise TypeError(
                    "operator.outputs must be a Operator.OperatorList "
                    "not %r." % type(operator.outputs))
            if any(not isinstance(i, Variable) for i in operator.inputs):
                raise TypeError(
                    "One input is not a Variable for operator %r - %r."
                    "" % (type(operator.raw_operator), operator))
            if any(not isinstance(i, Variable) for i in operator.outputs):
                raise TypeError(
                    "One output is not a Variable for operator %r - %r."
                    "" % (type(operator.raw_operator), operator))

        def _check_variable_in_(variable, operator):
            idop = id(operator)
            ids = set(id(op) for op in variable.operators_inputs_)
            if idop not in ids:
                raise RuntimeError(
                    "Operator %r not registered in the list of operators "
                    "of %r taking it as an input [\n%s]." % (
                        operator, variable,
                        "\n".join(map(str, variable.operators_inputs_))))

        def _check_variable_out_(variable, operator):
            if variable.is_fed:
                add = ["", "--DEBUG-INFO--"]
                for scope in self.scopes:
                    add.append('---')
                    add.append(pprint.pformat(
                        scope.variable_name_mapping))
                    add.append('---')
                    for var in scope.variables.values():
                        add.append("   is_fed=%s %s - n_in=%d n_out=%d" % (
                            getattr(var, 'is_fed', '?'), var,
                            len(var.operators_inputs_),
                            len(var.operators_outputs_)))
                    add.append('---')
                    for op in scope.operators.values():
                        add.append("   is_evaluated=%s %s" % (
                            getattr(op, 'is_evaluated', '?'), op))
                add.append('---')
                for v in operator.inputs:
                    add.append(" inputs={}".format(v))
                for v in operator.outputs:
                    add.append(" outputs={}".format(v))
                add.append('--- operator producing this variable--')
                for op in variable.operators_outputs_:
                    add.append(str(op))
                raise RuntimeError(
                    "A variable is already assigned ({}) "
                    "for operator '{}' (name='{}'). "
                    "operator.is_evaluated={}, inputs.is_fed={}, "
                    "outputs.is_fed={}. "
                    "This may still happen if a converter is a "
                    "combination of sub-estimators and one "
                    "of them is producing this output. "
                    "In that case, an identity node must be "
                    "added.{}".format(
                        variable, operator.type,
                        operator.onnx_name, operator.is_evaluated,
                        [v.is_fed for v in operator.inputs],
                        [v.is_fed for v in operator.outputs],
                        "\n".join(add)))

        if verbose > 0:
            print("[convert_operators] begin")
        self._initialize_graph_status_for_traversing()
        fed_variables = {i.name: i for i in container.initializers}
        changes = 1
        n_iter = 0
        while changes > 0:
            n_iter += 1
            changes = 0
            ops = list(self.unordered_operator_iterator())
            if verbose > 0:
                print("[convert_operators] iteration %d - n_vars=%d "
                      "n_ops=%d" % (
                          n_iter, len(fed_variables), len(ops)))
            for operator in ops:
                _check_operator_(operator)
                for var in operator.inputs:
                    if var.is_fed:
                        fed_variables[var.onnx_name] = var
                if (all(variable.is_fed for variable in operator.inputs) and
                        not operator.is_evaluated):

                    for variable in operator.inputs:
                        _check_variable_in_(variable, operator)
                    for variable in operator.outputs:
                        _check_variable_out_(variable, operator)

                    self.call_shape_calculator(operator)
                    self.call_converter(operator, container, verbose=verbose)

                    # If an operator contains a sequence of operators,
                    # output variables are not necessarily known at this stage.
                    operator.init_status(is_evaluated=True)
                    for variable in operator.outputs:
                        if all(op.is_evaluated
                               for op in variable.operators_outputs_):
                            variable.init_status(is_fed=True)
                            fed_variables[variable.onnx_name] = variable
                    fed_variables.update(
                        {i.name: i for i in container.initializers
                         if i.name not in fed_variables})
                    self._propagate_status(operator, container, fed_variables,
                                           verbose=verbose)

                    # unfed some variables (it happens when a node
                    # shares an output with another node)
                    rem = []
                    for n, var in fed_variables.items():
                        if not hasattr(var, 'operators_outputs_'):
                            # initializer
                            continue
                        if any(not o.is_evaluated
                               for o in var.operators_outputs_):
                            rem.append(n)
                    for r in rem:
                        v = fed_variables[r]
                        v.init_status(is_fed=False)
                        del fed_variables[v.onnx_name]
                    changes += 1

            if verbose > 0:
                print("[convert_operators] end iter: %d - n_vars=%d" % (
                    n_iter, len(fed_variables)))
        if verbose > 0:
            print("[convert_operators] end.")

        # Last verification.
        not_evaluated = []
        for op in self.unordered_operator_iterator():
            if not op.is_evaluated:
                not_evaluated.append(op)
        if len(not_evaluated) > 0:
            rows = ["---VARS---"]
            for var in self.unordered_variable_iterator():
                rows.append(
                    "is_fed=%r is_leaf=%r is_root=%r - %r - n_in=%d n_out=%d"
                    "" % (var.is_fed, var.is_leaf, var.is_root, var,
                          len(var.operators_inputs_),
                          len(var.operators_outputs_)))
            rows.append("---OPERATORS---")
            for op in self.unordered_operator_iterator():
                rows.append("is_eval=%r - %r" % (op.is_evaluated, op))
            rows.append("---NODES---")
            for node in container.nodes:
                rows.append("%s: %r -> %r" % (
                    node.op_type, node.input, node.output))
            raise RuntimeError(
                "Not all operators have been evaluated. A variable name "
                "is probably misspelled.\n%s"
                "" % "\n".join(rows))

        # Input and output
        if len(self.scopes[0].input_variables) > 0:
            inputs = self.scopes[0].input_variables
        else:
            inputs = [v for v in self.unordered_variable_iterator()
                      if v.is_root]
        for i in inputs:
            container.add_input(i)
        outputs = [v for v in self.unordered_variable_iterator()
                   if v.is_leaf]

        # The function checks that for output variable,
        # raw_name equal onnx_name. It swaps names if it is not the case.
        to_swap = []
        for out in outputs:
            if out.raw_name != out.onnx_name:
                to_swap.append(out)
        if len(to_swap) != 0:
            swaped = set()
            for var in to_swap:
                if var.raw_name in swaped:
                    continue
                swaped.add(var.raw_name)
                if verbose > 1:
                    print("[convert_operators] %r <-> %r." % (
                        var.raw_name, var.onnx_name))
                old_name = var.onnx_name
                new_name = var.raw_name

                try:
                    container.swap_names(old_name, new_name)
                except NotImplementedError as e:
                    logger.debug(
                        '[Topo] unable to swap %r and %r (%r).',
                        old_name, new_name, e)
                    continue

                for v in self.unordered_variable_iterator():
                    if v.onnx_name == old_name:
                        v.set_onnx_name(new_name)
                    elif v.onnx_name == new_name:
                        v.set_onnx_name(old_name)

        for o in outputs:
            container.add_output(o)


def convert_topology(topology, model_name, doc_string, target_opset,
                     channel_first_inputs=None,
                     options=None, remove_identity=True,
                     verbose=0):
    """
    This function is used to convert our Topology object defined in
    _parser.py into a ONNX model (type: ModelProto).

    :param topology: The Topology object we are going to convert
    :param model_name: GraphProto's name. Let "model" denote the
        returned model. The string "model_name" would be
        assigned to "model.graph.name."
    :param doc_string: A string attached to the produced model
    :param target_opset: number or dictionary,
        for example, 7 for ONNX 1.2, and 8 for ONNX 1.3,
        a dictionary is used to indicate different opset for
        different domains
    :param options: see :ref:`l-conv-options`
    :param remove_identity: removes identity nodes
        include '1.1.2', '1.2', and so on.
    :param verbose: displays information while converting
    :return: a ONNX ModelProto
    """
    if target_opset is None:
        target_opset = get_latest_tested_opset_version()
    if isinstance(target_opset, dict):
        onnx_target_opset = target_opset.get(
            '', get_latest_tested_opset_version())
    else:
        onnx_target_opset = target_opset
    if onnx_target_opset > get_opset_number_from_onnx():
        found = get_opset_number_from_onnx()
        raise RuntimeError(
            "Parameter target_opset {} > {} is higher than the "
            "version of the installed onnx package. See "
            "https://github.com/onnx/onnx/blob/master/docs/"
            "Versioning.md#released-versions"
            ".".format(onnx_target_opset, found))
    if onnx_target_opset > get_latest_tested_opset_version():
        warnings.warn(
            "Parameter target_opset {} > {} is higher than the "
            "the latest tested version"
            ".".format(
                onnx_target_opset,
                get_latest_tested_opset_version()))

    container = ModelComponentContainer(
        target_opset, options=options,
        registered_models=topology.registered_models,
        white_op=topology.raw_model._white_op,
        black_op=topology.raw_model._black_op,
        verbose=verbose)

    # Traverse the graph from roots to leaves
    # This loop could eventually be parallelized.
    topology.convert_operators(container=container, verbose=verbose)
    container.ensure_topological_order()

    if len(container.inputs) == 0:
        raise RuntimeError("No detected inputs after conversion.")
    if len(container.outputs) == 0:
        raise RuntimeError("No detected outputs after conversion.")
    if verbose >= 2:
        print("---NODES---")
        for node in container.nodes:
            print("  %s - %s: %r -> %r" % (
                node.op_type, node.name, node.input, node.output))

    # Create a graph from its main components
    if container.target_opset_onnx < 9:
        # When calling ModelComponentContainer's add_initializer(...),
        # nothing is added into the input list. However, for ONNX target
        # opset < 9, initializers should also be a part of model's
        # (GraphProto) inputs. Thus, we create ValueInfoProto objects
        # from initializers (type: TensorProto) directly and then add
        # them into model's input list.
        extra_inputs = []  # ValueInfoProto list of the initializers
        for tensor in container.initializers:
            # Sometimes (especially when creating optional input values
            # such as RNN's initial hidden state), an initializer is also
            # one of the original model's input, so it has been added into
            # the container's input list. If this is the case, we need to
            # skip one iteration to avoid duplicated inputs.
            if tensor.name in [value_info.name for value_info in
                               container.inputs]:
                continue

            # Initializers are always tensors so we can just call
            # make_tensor_value_info(...).
            value_info = make_tensor_value_info(
                tensor.name, tensor.data_type, tensor.dims)
            extra_inputs.append(value_info)

        # Before ONNX opset 9, initializers were needed to be passed in
        # with inputs.
        graph = make_graph(container.nodes, model_name,
                           container.inputs + extra_inputs,
                           container.outputs, container.initializers)
    else:
        # In ONNX opset 9 and above, initializers are included as
        # operator inputs and therefore do not need to be passed as
        # extra_inputs.
        graph = make_graph(
            container.nodes, model_name, container.inputs,
            container.outputs, container.initializers)

    # Add extra information related to the graph
    graph.value_info.extend(container.value_info)

    # Create model
    onnx_model = make_model(graph)

    # Update domain version
    opv = min(onnx_target_opset,
              _get_main_opset_version(onnx_model) or onnx_target_opset)
    if not _update_domain_version(container, onnx_model, verbose=verbose):
        # Main opset was not added. Doing it here.
        op_set = onnx_model.opset_import.add()
        op_set.domain = ''
        op_set.version = opv
        if verbose > 0:
            print('[convert_topology] +opset: name=%r, version=%s' % (
                '', opv))

    # Add extra information
    irv = OPSET_TO_IR_VERSION.get(opv, onnx_proto.IR_VERSION)
    onnx_model.ir_version = irv
    onnx_model.producer_name = utils.get_producer()
    onnx_model.producer_version = utils.get_producer_version()
    onnx_model.domain = utils.get_domain()
    onnx_model.model_version = utils.get_model_version()
    onnx_model.doc_string = doc_string

    # Removes many identity nodes,
    # the converter may introduct identity nodes
    # after a zipmap operator and onnx <= 1.7 does not
    # support that. It does not use onnxconverter-common
    # as the optimizer only support opset >= 9.
    if remove_identity:
        onnx_model = onnx_remove_node_identity(onnx_model)

    return onnx_model


def _update_domain_version(container, onnx_model, verbose=0):
    # Merge operator sets for the same domain, the largest version
    # number would be kept
    purified_operator_set = dict()
    for op_domain, op_version in container.node_domain_version_pair_sets:
        if op_domain not in purified_operator_set:
            purified_operator_set[op_domain] = op_version
        else:
            purified_operator_set[op_domain] = max(
                purified_operator_set[op_domain], op_version)

    # Fill operator sets
    i = 0
    for op_domain, op_version in purified_operator_set.items():
        if op_version is None:
            continue
        if i == 0 and len(onnx_model.opset_import) == 1:
            # Overwrite the default operator set created by
            # make_model(...)
            op_set = onnx_model.opset_import[0]
        else:
            # Just create one ONNX element in opset_import
            op_set = onnx_model.opset_import.add()
        if verbose > 0:
            print('[_update_domain_version] +opset %d: name=%r, version=%s' % (
                i, op_domain, op_version))
        op_set.domain = op_domain
        if op_set != '':
            max_supported = get_default_opset_for_domain(op_domain)
            if max_supported is not None and max_supported < op_version:
                raise RuntimeError(
                    "The model is using version %d of domain %r not supported "
                    "yet by this library. You need to specify "
                    "target_opset={%r: %r}." % (
                        op_version, op_domain, op_domain, max_supported))
        op_set.version = op_version

        i += 1
        if container.target_opset_any_domain(op_domain) < op_version:
            raise RuntimeError(
                'The specified opset %d is too low to convert '
                'this model, which requires at least opset '
                '%d.' % (
                    container.target_opset_any_domain(op_domain),
                    op_version))
    return '' in purified_operator_set


def _get_main_opset_version(model):
    """
    Returns the main opset version.
    """
    mld = None
    for op in model.opset_import:
        if op.domain == '':
            return op.version
        if op.domain == "ai.onnx.ml":
            mld = op.version
    if mld is not None:
        return OPSET_ML_TO_OPSET.get(mld, None)
    return None



# SPDX-License-Identifier: Apache-2.0


# To register a converter for scikit-learn operators,
# import associated modules here.
from . import ada_boost
from . import array_feature_extractor
from . import bagging
from . import binariser
from . import calibrated_classifier_cv
from . import cast_op
from . import class_labels
from . import concat_op
from . import cross_decomposition
from . import decision_tree
from . import decomposition
from . import dict_vectoriser
from . import feature_hasher
from . import feature_selection
from . import flatten_op
from . import function_transformer
from . import gamma_regressor
from . import gaussian_process
from . import gaussian_mixture
from . import gradient_boosting
from . import grid_search_cv
from . import id_op
from . import imputer_op
from . import isolation_forest
from . import kernel_pca
from . import k_bins_discretiser
from . import k_means
from . import label_binariser
from . import label_encoder
from . import linear_classifier
from . import linear_regressor
from . import local_outlier_factor
from . import multilayer_perceptron
from . import multioutput
from . import multiply_op
from . import naive_bayes
from . import nearest_neighbours
from . import normaliser
from . import one_hot_encoder
from . import one_vs_one_classifier
from . import one_vs_rest_classifier
from . import ordinal_encoder
from . import ovr_decision_function
from . import pipelines
from . import polynomial_features
from . import power_transformer
from . import quadratic_discriminant_analysis
from . import random_forest
from . import random_projection
from . import random_trees_embedding
from . import ransac_regressor
from . import replace_op
from . import scaler_op
from . import sequence
from . import sgd_classifier
from . import sgd_oneclass_svm
from . import stacking
from . import support_vector_machines
from . import text_vectoriser
from . import tfidf_transformer
from . import tfidf_vectoriser
from . import voting_classifier
from . import voting_regressor
from . import zip_map

__all__ = [
    ada_boost,
    array_feature_extractor,
    bagging,
    binariser,
    calibrated_classifier_cv,
    cast_op,
    class_labels,
    concat_op,
    cross_decomposition,
    decision_tree,
    decomposition,
    dict_vectoriser,
    feature_hasher,
    feature_selection,
    flatten_op,
    function_transformer,
    gamma_regressor,
    gaussian_process,
    gaussian_mixture,
    gradient_boosting,
    grid_search_cv,
    id_op,
    imputer_op,
    isolation_forest,
    kernel_pca,
    k_bins_discretiser,
    k_means,
    label_binariser,
    label_encoder,
    linear_classifier,
    linear_regressor,
    local_outlier_factor,
    multilayer_perceptron,
    multioutput,
    multiply_op,
    naive_bayes,
    nearest_neighbours,
    normaliser,
    one_hot_encoder,
    one_vs_one_classifier,
    one_vs_rest_classifier,
    ordinal_encoder,
    ovr_decision_function,
    pipelines,
    polynomial_features,
    power_transformer,
    quadratic_discriminant_analysis,
    random_forest,
    random_projection,
    random_trees_embedding,
    ransac_regressor,
    replace_op,
    scaler_op,
    sequence,
    sgd_classifier,
    sgd_oneclass_svm,
    stacking,
    support_vector_machines,
    text_vectoriser,
    tfidf_transformer,
    tfidf_vectoriser,
    voting_classifier,
    voting_regressor,
    zip_map,
]



"""
@file
@brief Overloads :epkg:`TfidfVectorizer` and :epkg:`CountVectorizer`.
"""
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
try:
    from sklearn.feature_extraction.text import (
        _VectorizerMixin as VectorizerMixin)
except ImportError:  # pragma: no cover
    # scikit-learn < 0.23
    from sklearn.feature_extraction.text import VectorizerMixin


class NGramsMixin(VectorizerMixin):
    """
    Overloads method `_word_ngrams
    <https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/feature_extraction/text.py#L148>`_
    to get tuples instead of string in member `vocabulary_
    <https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.CountVectorizer.html>`_.
    of `TfidfVectorizer` or :epkg:`CountVectorizer`.
    It contains the list of n-grams used to process documents.
    See :class:`TraceableCountVectorizer` and :class:`TraceableTfidfVectorizer`
    for example.
    """

    def _word_ngrams(self, tokens, stop_words=None):
        """Turn tokens into a sequence of n-grams after stop words filtering"""
        # handle stop words
        if stop_words is not None:
            tokens = [w for w in tokens if w not in stop_words]

        if tokens is not None:
            new_tokens = []
            for token in tokens:
                val = (token,) if isinstance(token, str) else token
                if not isinstance(val, tuple):
                    raise TypeError(
                        f"Unexpected type {type(val)}:{val!r} for a token.")
                if any(map(lambda x: not isinstance(x, str), val)):
                    raise TypeError(
                        f"Unexpected type {val!r}, one part of a "
                        f"token is not a string.")
                new_tokens.append(val)
            tokens = new_tokens

        # handle token n-grams
        min_n, max_n = self.ngram_range
        if max_n != 1:
            original_tokens = tokens
            if min_n == 1:
                # no need to do any slicing for unigrams
                # just iterate through the original tokens
                tokens = list(original_tokens)
                min_n += 1
            else:
                tokens = []

            n_original_tokens = len(original_tokens)

            # bind method outside of loop to reduce overhead
            tokens_append = tokens.append

            def space_join(tokens):
                new_tokens = []
                for token in tokens:
                    if isinstance(token, str):
                        new_tokens.append(token)
                    elif isinstance(token, tuple):
                        new_tokens.extend(token)
                    else:
                        raise TypeError(  # pragma: no cover
                            f"Unable to build a n-grams out of {tokens}.")
                return tuple(new_tokens)

            for n in range(min_n,
                           min(max_n + 1, n_original_tokens + 1)):
                for i in range(n_original_tokens - n + 1):
                    tokens_append(space_join(original_tokens[i: i + n]))
        return tokens

    @staticmethod
    def _fix_vocabulary(expected, new_voc):
        update = {}
        for w, wid in new_voc.items():
            if not isinstance(w, tuple):
                raise TypeError(
                    f"Tuple is expected for a token not {type(w)}.")
            s = " ".join(w)
            if s in expected:
                if expected[s] != wid:
                    update[w] = wid
        if update:
            new_voc.update(update)
        duplicates = {}
        for w, wid in new_voc.items():
            if wid not in duplicates:
                duplicates[wid] = {w}
            else:
                duplicates[wid].add(w)
        dups = {k: v for k, v in duplicates.items() if len(v) > 1}
        return update, dups


class TraceableCountVectorizer(CountVectorizer, NGramsMixin):
    """
    Inherits from :class:`NGramsMixin` which overloads method `_word_ngrams
    <https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/feature_extraction/text.py#L148>`_
    to keep more information about n-grams but still produces the same
    outputs than `CountVectorizer`.

    .. runpython::
        :showcode:

        import numpy
        from sklearn.feature_extraction.text import CountVectorizer
        from mlinsights.mlmodel.sklearn_text import TraceableCountVectorizer
        from pprint import pformat

        corpus = numpy.array([
            "This is the first document.",
            "This document is the second document.",
            "Is this the first document?",
            "",
        ]).reshape((4, ))

        print('CountVectorizer from scikit-learn')
        mod1 = CountVectorizer(ngram_range=(1, 2))
        mod1.fit(corpus)
        print(mod1.transform(corpus).todense()[:2])
        print(pformat(mod1.vocabulary_)[:100])

        print('TraceableCountVectorizer from scikit-learn')
        mod2 = TraceableCountVectorizer(ngram_range=(1, 2))
        mod2.fit(corpus)
        print(mod2.transform(corpus).todense()[:2])
        print(pformat(mod2.vocabulary_)[:100])

    A weirder example with
    @see cl TraceableTfidfVectorizer shows more differences.

    The class is training an instance of CountVectorizer on the
    same data. This is used to update the vocabulary to match
    the same columns as the one obtained with scikit-learn.
    scikit-learn cannot distinguish between bi gram ("a b", "c") and
    ("a", "b c"). Therefore, there are merged into the same
    column by scikit-learn. This class, even if it is able to distinguish
    between them, keeps the same ambiguity.
    """

    def _word_ngrams(self, tokens, stop_words=None):
        return NGramsMixin._word_ngrams(
            self, tokens=tokens, stop_words=stop_words)

    def fit(self, X, y=None):
        # scikit-learn implements fit_transform and fit calls it.
        new_self = CountVectorizer(**self.get_params())
        new_self._word_ngrams = self._word_ngrams
        new_self.fit(X, y=y)
        for k, v in new_self.__dict__.items():
            if k.endswith("_") and not k.endswith("__"):
                setattr(self, k, v)
        same = CountVectorizer(**self.get_params())
        same.fit(X, y=y)
        self.same_ = same
        if self.stop_words != same.stop_words:
            raise AssertionError(
                f"Different stop_words {self.stop_words} "
                f"!= {same.stop_words}.")
        update, dups = self._fix_vocabulary(same.vocabulary_, self.vocabulary_)
        self.updated_vocabulary_ = update
        self.duplicated_vocabulary_ = dups
        return self

    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X)


class TraceableTfidfVectorizer(TfidfVectorizer, NGramsMixin):
    """
    Inherits from :class:`NGramsMixin` which overloads method `_word_ngrams
    <https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/feature_extraction/text.py#L148>`_
    to keep more information about n-grams but still produces the same
    outputs than `TfidfVectorizer`.

    .. runpython::
        :showcode:

        import numpy
        from sklearn.feature_extraction.text import TfidfVectorizer
        from mlinsights.mlmodel.sklearn_text import TraceableTfidfVectorizer
        from pprint import pformat

        corpus = numpy.array([
            "This is the first document.",
            "This document is the second document.",
            "Is this the first document?",
            "",
        ]).reshape((4, ))

        print('TfidfVectorizer from scikit-learn')
        mod1 = TfidfVectorizer(ngram_range=(1, 2),
                               token_pattern="[a-zA-Z ]{1,4}")
        mod1.fit(corpus)
        print(mod1.transform(corpus).todense()[:2])
        print(pformat(mod1.vocabulary_)[:100])

        print('TraceableTfidfVectorizer from scikit-learn')
        mod2 = TraceableTfidfVectorizer(ngram_range=(1, 2),
                                       token_pattern="[a-zA-Z ]{1,4}")
        mod2.fit(corpus)
        print(mod2.transform(corpus).todense()[:2])
        print(pformat(mod2.vocabulary_)[:100])

    The class is training an instance of TfidfVectorizer on the
    same data. This is used to update the vocabulary to match
    the same columns as the one obtained with scikit-learn.
    scikit-learn cannot distinguish between bi gram ("a b", "c") and
    ("a", "b c"). Therefore, there are merged into the same
    column by scikit-learn. This class, even if it is able to distinguish
    between them, keeps the same ambiguity.    """

    def _word_ngrams(self, tokens, stop_words=None):
        return NGramsMixin._word_ngrams(
            self, tokens=tokens, stop_words=stop_words)

    def fit(self, X, y=None):
        super().fit(X, y=y)
        same = TfidfVectorizer(**self.get_params())
        same.fit(X, y=y)
        self.same_ = same
        if self.stop_words != same.stop_words:
            raise AssertionError(
                f"Different stop_words {self.stop_words} "
                f"!= {same.stop_words}.")
        update, dups = self._fix_vocabulary(same.vocabulary_, self.vocabulary_)
        self.updated_vocabulary_ = update
        self.duplicated_vocabulary_ = dups
        return self

    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X)



# SPDX-License-Identifier: Apache-2.0

"""
Optimization of :epkg:`ONNX` graphs.
Functions in *onnxconverter-common* do not support
opset < 9.
"""
from logging import getLogger
from onnx.helper import make_graph
from ._onnx_optimisation_common import (
    _rename_node_input, _rename_node_output,
    _apply_optimisation_on_graph, _apply_remove_node_fct_node)


logger = getLogger('skl2onnx')


def onnx_remove_node_identity(onnx_model, recursive=True, debug_info=None):
    """
    Removes as many *Identity* nodes as possible.
    The function looks into every node and subgraphs if
    *recursive* is True for identity node. Unless such a
    node directy connects one input to one output, it will
    be removed and every other node gets its inputs or
    outputs accordingly renamed.

    :param onnx_model: onnx model
    :param recursive: looks into subgraphs
    :param debug_info: debug information (private)
    :return: new onnx model
    """
    if debug_info is None:
        debug_info = [str(type(onnx_model)).split('.')[-1].strip("'>")]
    else:
        debug_info = debug_info + \
            [str(type(onnx_model)).split('.')[-1].strip("'>")]

    if hasattr(onnx_model, 'graph'):
        return _apply_optimisation_on_graph(
            onnx_remove_node_identity, onnx_model,
            recursive=recursive, debug_info=debug_info)

    graph = onnx_model

    inputs = set(i.name for i in graph.input)
    outputs = set(o.name for o in graph.output)

    def retrieve_idnodes(graph, existing_nodes):
        idnodes = []
        for i, exnode in enumerate(existing_nodes):
            if exnode is None:
                continue
            if exnode.op_type == 'Identity':
                input = exnode.input[0]
                output = exnode.output[0]
                idnodes.append((i, exnode, input, output))
        return idnodes

    def retrieve_local_variables_subgraphs(graph):
        local = set()
        existing = set(i.name for i in graph.input)
        for node in graph.node:
            for i in node.input:
                if i not in existing:
                    local.add(i)
            for o in node.output:
                existing.add(o)
            res = retrieve_local_variables_nodes([node])
            for r in res:
                if r not in existing:
                    local.add(r)
        return local

    def retrieve_local_variables_nodes(nodes):
        names = set()
        for node in nodes:
            for att in node.attribute:
                if att.g:
                    names |= retrieve_local_variables_subgraphs(att.g)
        return names

    nodes = list(graph.node)
    local_variables = retrieve_local_variables_nodes(nodes)
    rem = 1
    while rem > 0:
        rem = 0
        idnodes = retrieve_idnodes(graph, nodes)
        restart = False
        for i, _, inp, out in idnodes:
            if restart:
                break
            if nodes[i] is None:
                # Already removed.
                continue
            if inp in inputs and out in outputs:
                # Cannot be removed.
                continue
            if out in local_variables:
                # out is used a local variable, this case is not implemented
                continue
            if not restart and out not in outputs:
                # We cannot change an output name.
                for j in range(len(nodes)):
                    if nodes[j] is None:
                        continue
                    if out in nodes[j].input:
                        nodes[j] = _rename_node_input(nodes[j], out, inp)
                        logger.debug(
                            '[VarId-a] rename node input %r into %r' % (
                                out, inp))
                        rem += 1
                        if nodes[j].op_type == 'Identity':
                            restart = True
                logger.debug('[NodeId-a] remove %r' % nodes[i])
                nodes[i] = None
                rem += 1
                continue
            if (not restart and inp not in inputs and inp not in outputs and
                    out not in outputs):
                # We cannot change an input name or an output name.
                for j in range(len(nodes)):
                    if nodes[j] is None:
                        continue
                    if inp in nodes[j].output:
                        nodes[j] = _rename_node_output(nodes[j], inp, out)
                        logger.debug(
                            '[Var] rename node output %r into %r' % (
                                out, inp))
                        rem += 1
                        if nodes[j].op_type == 'Identity':
                            restart = True
                    if inp in nodes[j].input:
                        nodes[j] = _rename_node_input(nodes[j], inp, out)
                        logger.debug(
                            '[VarId-b] rename node input %r into %r' % (
                                out, inp))
                        rem += 1
                        if nodes[j].op_type == 'Identity':
                            restart = True
                logger.debug('[NodeId-b] remove %r' % nodes[i])
                nodes[i] = None
                rem += 1

    if recursive:
        # Handles subgraphs.
        for i in range(len(nodes)):
            node = nodes[i]
            if node is None or not (node.attribute):
                continue
            nodes[i] = _apply_remove_node_fct_node(
                onnx_remove_node_identity,
                node, recursive=True, debug_info=debug_info + [node.name])

    # Finally create the new graph.
    nodes = list(filter(lambda n: n is not None, nodes))
    graph = make_graph(nodes, onnx_model.name,
                       onnx_model.input, onnx_model.output,
                       onnx_model.initializer)

    graph.value_info.extend(onnx_model.value_info)
    return graph



# SPDX-License-Identifier: Apache-2.0


from ..proto import onnx_proto
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_feature_selection(scope: Scope, operator: Operator,
                                      container: ModelComponentContainer):
    op = operator.raw_operator
    # Get indices of the features selected
    index = op.get_support(indices=True)
    if len(index) == 0:
        raise RuntimeError(
            "Model '{}' did not select any feature. "
            "This model cannot be converted into ONNX."
            "".format(op.__class__.__name__))
    output_name = operator.outputs[0].full_name
    if index.any():
        column_indices_name = scope.get_unique_variable_name('column_indices')

        container.add_initializer(column_indices_name,
                                  onnx_proto.TensorProto.INT64,
                                  [len(index)], index)

        container.add_node(
            'ArrayFeatureExtractor',
            [operator.inputs[0].full_name, column_indices_name],
            output_name, op_domain='ai.onnx.ml',
            name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
    else:
        container.add_node('ConstantOfShape', operator.inputs[0].full_name,
                           output_name, op_version=9)


register_converter('SklearnGenericUnivariateSelect',
                   convert_sklearn_feature_selection)
register_converter('SklearnRFE', convert_sklearn_feature_selection)
register_converter('SklearnRFECV', convert_sklearn_feature_selection)
register_converter('SklearnSelectFdr', convert_sklearn_feature_selection)
register_converter('SklearnSelectFpr', convert_sklearn_feature_selection)
register_converter('SklearnSelectFromModel', convert_sklearn_feature_selection)
register_converter('SklearnSelectFwe', convert_sklearn_feature_selection)
register_converter('SklearnSelectKBest', convert_sklearn_feature_selection)
register_converter('SklearnSelectPercentile',
                   convert_sklearn_feature_selection)
register_converter('SklearnVarianceThreshold',
                   convert_sklearn_feature_selection)



# SPDX-License-Identifier: Apache-2.0

import copy
from collections import OrderedDict
import warnings
from sklearn.base import BaseEstimator
from sklearn.pipeline import Pipeline


def enumerate_model_names(model, prefix="", short=True):
    """
    Enumerates ``tuple (name, model)`` associated
    to the model itself.
    """
    if isinstance(model, (list, tuple)):
        if all(map(lambda x: isinstance(x, tuple) and len(x) in (2, 3),
                   model)):
            for i, named_mod in enumerate(model):
                name, mod = named_mod[:2]
                p = (name if short and prefix == ""
                     else "{}__{}".format(prefix, name))
                for t in enumerate_model_names(mod, p, short=short):
                    yield t
        else:
            for i, mod in enumerate(model):
                p = (i if short and prefix == ""
                     else "{}__{}".format(prefix, i))
                for t in enumerate_model_names(mod, p, short=short):
                    yield t
    elif isinstance(model, (dict, OrderedDict)):
        for name, mod in model.items():
            p = (name if short and prefix == ""
                 else "{}__{}".format(prefix, name))
            for t in enumerate_model_names(mod, p, short=short):
                yield t
    else:
        yield (prefix, model)
        reserved_atts = {'transformers', 'steps', 'transformer_list',
                         'named_estimators_', 'named_transformers_',
                         'transformer_', 'estimator_'}
        for key in dir(model):
            if (key in ('estimators_', 'estimator') and
                    hasattr(model, 'named_estimators_')):
                continue
            if (key in ('transformers_', 'transformers') and
                    hasattr(model, 'named_transformers_')):
                continue
            if (key in reserved_atts or
                    (key.endswith("_") and not key.endswith("__") and
                     not key.startswith('_'))):
                try:
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore", FutureWarning)
                        obj = getattr(model, key)
                except AttributeError:
                    continue
                if (hasattr(obj, 'get_params') and
                        isinstance(obj, BaseEstimator)):
                    prefix = (key if short and prefix == ""
                              else "{}__{}".format(prefix, key))
                    yield (prefix, obj)
                elif isinstance(obj, (list, tuple, dict, OrderedDict)):
                    if not short or key not in reserved_atts:
                        prefix = (key if short and prefix == ""
                                  else "{}__{}".format(prefix, key))
                    for t in enumerate_model_names(obj, prefix, short=short):
                        yield t


def has_pipeline(model):
    """
    Tells if a model contains a pipeline.
    """
    return any(map(lambda x: isinstance(x[1], Pipeline),
                   enumerate_model_names(model)))


def _process_options(model, options):
    """
    Converts options defined as string into options
    ``id(model): options``. The second format is not
    pickable.
    """
    if options is None:
        return None
    if all(map(lambda x: not isinstance(x, str), options)):
        # No need to transform.
        return _process_pipeline_options(model, options)

    new_options = copy.deepcopy(options)
    names = dict(enumerate_model_names(model))
    for k, v in options.items():
        if k in names:
            new_options[id(names[k])] = v
            continue
        try:
            ri = k.rindex('__')
            m2, k2 = k[:ri], k[ri + 2:]
        except ValueError:
            key = id(model)
            if key not in new_options:
                new_options[key] = {}
            new_options[key][k] = v
            continue
        if m2 in names:
            key = id(names[m2])
            if key not in new_options:
                new_options[key] = {}
            new_options[key][k2] = v
            continue
        raise RuntimeError(
            "Unable to find model name '{}' or '{}' in \n{}".format(
                k, m2, list(sorted(names))))

    return _process_pipeline_options(model, new_options)


def _process_pipeline_options(model, options):
    """
    Tells the final classifier of a pipeline that
    options 'zipmap', 'nocl' or 'output_class_labels'
    were attached to the pipeline.
    """
    new_options = None
    names = dict(enumerate_model_names(model))
    for k, v in names.items():
        if id(v) in options and isinstance(v, Pipeline):
            opts = options[id(v)]
            last = v.steps[-1][1]
            key = id(last)
            for opt, val in opts.items():
                if opt not in {'zipmap', 'nocl', 'output_class_labels'}:
                    continue
                if new_options is None:
                    new_options = copy.deepcopy(options)
                if key not in new_options:
                    new_options[key] = {}
                if opt not in new_options[key]:
                    new_options[key][opt] = val
    return new_options or options



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..common._registration import register_shape_calculator
from ..common.data_types import FloatTensorType, Int64TensorType


def calculate_sklearn_random_trees_embedding_output_shapes(operator):
    op = operator.raw_operator.one_hot_encoder_
    categories_len = 0
    for index, categories in enumerate(op.categories_):
        if hasattr(op, 'drop_idx_') and op.drop_idx_ is not None:
            categories = (categories[np.arange(len(categories)) !=
                          op.drop_idx_[index]])
        categories_len += len(categories)
    instances = operator.inputs[0].get_first_dimension()
    if np.issubdtype(op.dtype, np.signedinteger):
        operator.outputs[0].type = Int64TensorType([instances, categories_len])
    else:
        operator.outputs[0].type = FloatTensorType([instances, categories_len])


register_shape_calculator(
    'SklearnRandomTreesEmbedding',
    calculate_sklearn_random_trees_embedding_output_shapes)



# SPDX-License-Identifier: Apache-2.0
import numpy as np
from ..proto import onnx_proto
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_class_labels(scope: Scope, operator: Operator,
                                 container: ModelComponentContainer):
    if getattr(operator, 'is_multi_output', False):
        classes = operator.classes
        if not isinstance(classes, list):
            raise RuntimeError(
                "classes must be a list of numpy arrays but is %r."
                "" % type(classes))
        names = []
        if classes[0].dtype in (np.int64, np.int32):
            for i, cl in enumerate(classes):
                cla = np.array(cl)
                name = scope.get_unique_variable_name(
                    operator.outputs[0].full_name + '_cst_%d' % i)
                container.add_initializer(
                    name, onnx_proto.TensorProto.INT64, list(cla.shape),
                    cla.tolist())
                names.append(name)
        else:
            for i, cl in enumerate(classes):
                name = scope.get_unique_variable_name(
                    operator.outputs[0].full_name + '_cst_%d' % i)
                clids = np.arange(len(cl), dtype=np.int64)
                container.add_initializer(
                    name, onnx_proto.TensorProto.INT64, list(clids.shape),
                    clids.tolist())
                namele = scope.get_unique_variable_name(
                    operator.outputs[0].full_name + '_le_%d' % i)
                container.add_node(
                    'LabelEncoder', name, namele, op_domain='ai.onnx.ml',
                    op_version=2, default_string='0', keys_int64s=clids,
                    values_strings=cl.tolist(),
                    name=scope.get_unique_operator_name(
                        'class_labels_le_%d' % i))
                names.append(namele)

        container.add_node(
            'SequenceConstruct', names, operator.outputs[0].full_name,
            name=scope.get_unique_operator_name('class_labels_seq'))
    else:
        classes = np.array(operator.classes)
        name = scope.get_unique_variable_name(
            operator.outputs[0].full_name + '_cst')

        if classes.dtype in (np.int64, np.int32):
            container.add_initializer(
                name, onnx_proto.TensorProto.INT64, list(classes.shape),
                classes.tolist())
        else:
            clids = np.arange(len(classes), dtype=np.int64)
            container.add_initializer(
                name, onnx_proto.TensorProto.INT64, list(clids.shape),
                clids.tolist())
            namele = scope.get_unique_variable_name(
                operator.outputs[0].full_name + '_le')
            container.add_node(
                'LabelEncoder', name, namele, op_domain='ai.onnx.ml',
                op_version=2, default_string='0', keys_int64s=clids,
                values_strings=classes.tolist(),
                name=scope.get_unique_operator_name('class_labels_le'))
            name = namele

        container.add_node(
            'Identity', name, operator.outputs[0].full_name)


register_converter(
    'SklearnClassLabels', convert_sklearn_class_labels)
